"""
Transformer nodes convert many inputs into a single output, or vice-versa.
"""

from __future__ import annotations

import itertools
from abc import ABC
from dataclasses import MISSING
from dataclasses import dataclass as as_dataclass
from typing import Any, ClassVar

from pandas import DataFrame
from pyiron_snippets.colors import SeabornColors
from pyiron_snippets.factory import classfactory

from pyiron_workflow import identifier
from pyiron_workflow.data import NOT_DATA, NotData
from pyiron_workflow.nodes.static_io import StaticNode


class Transformer(StaticNode, ABC):
    """
    Transformers are a special case of :class:`StaticNode` nodes that turn many inputs
    into a single output or vice-versa.
    """

    @property
    def color(self) -> str:
        """For drawing the graph"""
        return SeabornColors.blue


class InputsToList(Transformer, ABC):
    _length: ClassVar[int]  # Mandatory attribute for non-abstract subclasses
    _content_type_hint: ClassVar[object]
    _output_name: ClassVar[str] = "list"

    @classmethod
    def _build_outputs_preview(cls) -> dict[str, Any]:
        return {
            cls._output_name: (
                list
                if cls._content_type_hint is None
                else list[cls._content_type_hint]  # type: ignore[name-defined]
            )
        }

    @classmethod
    def _build_inputs_preview(cls) -> dict[str, tuple[Any, Any]]:
        return {
            f"item_{i}": (cls._content_type_hint, NOT_DATA) for i in range(cls._length)
        }

    @property
    def run_args(self) -> tuple[tuple, dict]:
        return (), self.inputs.to_value_dict()

    def _on_run(self, **inputs_to_value_dict):
        return list(inputs_to_value_dict.values())

    def process_run_result(self, run_output: Any | tuple) -> Any | tuple:
        self.outputs[self._output_name].value = run_output
        return run_output


@classfactory
def inputs_to_list_factory(
    n: int, class_name: str, use_cache: bool = True, content_type_hint: object = None, /
) -> type[InputsToList]:
    return (  # type: ignore[return-value]
        class_name,
        (InputsToList,),
        {
            "_length": n,
            "_content_type_hint": content_type_hint,
            "use_cache": use_cache,
        },
        {},
    )


def inputs_to_list(
    n: int,
    /,
    *node_args,
    use_cache: bool = True,
    content_type_hint: object = NOT_DATA,
    **node_kwargs,
):
    """
    Creates and returns an instance of a dynamically generated :class:`InputsToList`
        subclass with a specified number of inputs.

    Args:
        n (int): Number of input channels.
        use_cache (bool): Whether this node should default to caching its values.
            (Default is True.)
        content_type_hint (object): Type hint for the content of the list. Applies to
            all input channels and the content type of the returned list. (Default is
            NOT_DATA, which will leave the content type of the list unspecified.)
        *node_args: Positional arguments for the node instance.
        **node_kwargs: Keyword arguments for the node instance.

    Returns:
        InputsToList: An instance of the dynamically created :class:`InputsToList`
            subclass.
    """
    hint_identifier = (
        ""
        if content_type_hint is NOT_DATA
        else identifier.to_identifier(str(content_type_hint))
    )
    class_name = f"{InputsToList.__name__}{n}{hint_identifier}"
    inputs_to_list_factory.clear(class_name)
    cls = inputs_to_list_factory(
        n, class_name, use_cache, _parse_type_hint(content_type_hint)
    )
    cls.preview_io()
    return cls(*node_args, **node_kwargs)


def _parse_type_hint(type_hint: object) -> object:
    if type_hint is NOT_DATA:
        return None
    elif type_hint is None:
        return type(None)
    return type_hint


class ListToOutputs(Transformer, ABC):
    _length: ClassVar[int]  # Mandatory attribute for non-abstract subclasses
    _content_type_hint: ClassVar[object]
    _input_name: ClassVar[str] = "list"

    @classmethod
    def _build_outputs_preview(cls) -> dict[str, Any]:
        return {f"item_{i}": cls._content_type_hint for i in range(cls._length)}

    @classmethod
    def _build_inputs_preview(cls) -> dict[str, tuple[Any, Any]]:
        return {
            cls._input_name: (
                (
                    list
                    if cls._content_type_hint is None
                    else list[cls._content_type_hint]  # type: ignore[name-defined]
                ),
                NOT_DATA,
            )
        }

    @property
    def run_args(self) -> tuple[tuple, dict]:
        return (self.inputs[self._input_name].value,), {}

    def _on_run(self, input_object: list):
        return {f"item_{i}": v for i, v in enumerate(input_object)}

    def process_run_result(self, run_output: dict[str, Any]) -> dict[str, Any]:
        for k, v in run_output.items():
            self.outputs[k].value = v
        return run_output


@classfactory
def list_to_outputs_factory(
    n: int, class_name: str, use_cache: bool = True, content_type_hint: object = None, /
) -> type[ListToOutputs]:
    return (  # type: ignore[return-value]
        class_name,
        (ListToOutputs,),
        {
            "_length": n,
            "_content_type_hint": content_type_hint,
            "use_cache": use_cache,
        },
        {},
    )


def list_to_outputs(
    n: int,
    /,
    *node_args,
    use_cache: bool = True,
    content_type_hint: object = NOT_DATA,
    **node_kwargs,
) -> ListToOutputs:
    """
    Creates and returns an instance of a dynamically generated :class:`ListToOutputs`
    subclass with a specified number of outputs.

    Args:
        n (int): Number of output channels.
        use_cache (bool): Whether this node should default to caching its values.
            (Default is True.)
        content_type_hint (object): Type hint for the content of the list. Applies to
            all output channels and the content type of the provided list. (Default is
            NOT_DATA, which will leave the content type of the list unspecified.)
        *node_args: Positional arguments for the node instance.
        **node_kwargs: Keyword arguments for the node instance.

    Returns:
        ListToOutputs: An instance of the dynamically created :class:`ListToOutputs`
            subclass.
    """
    hint_identifier = (
        ""
        if content_type_hint is NOT_DATA
        else identifier.to_identifier(str(content_type_hint))
    )
    class_name = f"{ListToOutputs.__name__}{n}{hint_identifier}"
    list_to_outputs_factory.clear(class_name)
    cls = list_to_outputs_factory(
        n, class_name, use_cache, _parse_type_hint(content_type_hint)
    )
    cls.preview_io()
    return cls(*node_args, **node_kwargs)


class InputsToDict(Transformer, ABC):
    _output_name: ClassVar[str] = "dict"
    _input_specification: ClassVar[
        list[str] | dict[str, tuple[Any | None, Any | NotData]]
    ]

    @classmethod
    def _build_inputs_preview(cls) -> dict[str, tuple[Any | None, Any | NotData]]:
        if isinstance(cls._input_specification, list):
            return dict.fromkeys(cls._input_specification, (None, NOT_DATA))
        else:
            return cls._input_specification

    @classmethod
    def _build_outputs_preview(cls) -> dict[str, Any]:
        if isinstance(cls._input_specification, list):
            type_hint = dict[str, Any]
        else:
            first_type = next(iter(cls._input_specification.values()))[0] or Any
            type_hint = dict[str, first_type]  # type: ignore[misc, valid-type]
        return {cls._output_name: type_hint}

    @property
    def run_args(self) -> tuple[tuple, dict]:
        return (), self.inputs.to_value_dict()

    def _on_run(self, **inputs_to_value_dict):
        return inputs_to_value_dict

    def process_run_result(self, run_output: Any | tuple) -> Any | tuple:
        self.outputs[self._output_name].value = run_output
        return run_output

    @staticmethod
    def hash_specification(
        input_specification: list[str] | dict[str, tuple[Any | None, Any | NotData]],
    ):
        """For generating unique subclass names."""

        if isinstance(input_specification, list):
            return hash(tuple(input_specification))
        else:
            flattened_tuple = tuple(
                itertools.chain.from_iterable(
                    (key, *value) for key, value in input_specification.items()
                )
            )
            try:
                return hash(flattened_tuple)
            except Exception as e:
                raise ValueError(
                    f"To automatically generate a unique name for subclasses of "
                    f"{InputsToDict.__name__}, the input specification must be fully "
                    f"hashable, but it was not. Either pass fully hashable hints and "
                    f"defaults, or explicitly provide a class name suffix. Received "
                    f"specification: {input_specification}"
                ) from e


@classfactory
def inputs_to_dict_factory(
    input_specification: list[str] | dict[str, tuple[Any | None, Any | NotData]],
    class_name_suffix: str | None,
    use_cache: bool = True,
    /,
) -> type[InputsToDict]:
    if class_name_suffix is None:
        class_name_suffix = str(
            InputsToDict.hash_specification(input_specification)
        ).replace("-", "m")
    return (  # type: ignore[return-value]
        f"{InputsToDict.__name__}{class_name_suffix}",
        (InputsToDict,),
        {
            "_input_specification": input_specification,
            "use_cache": use_cache,
        },
        {},
    )


def inputs_to_dict(
    input_specification: list[str] | dict[str, tuple[Any | None, Any | NotData]],
    *node_args,
    class_name_suffix: str | None = None,
    use_cache: bool = True,
    **node_kwargs,
):
    """
    Build a new :class:`InputsToDict` subclass and instantiate it.

    Tries to automatically generate a subclass name by hashing the
    :param:`input_specification`. If such hashing fails, you will instead _need_ to
    provide an explicit :param:`class_name_suffix`

    Args:
        input_specification (list[str] | dict[str, tuple[Any | None, Any | NOT_DATA]]):
            The input channel names, or full input specification in the form
            `{key: (type_hint, default_value))}`.
        *node_args: Other args for the node instance.
        class_name_suffix (str | None): The suffix to use in the class name. (Default
            is None, try to generate the suffix by hashing :param:`input_specification`.
        use_cache (bool): Whether this node should default to caching its values.
            (Default is True.)
        **node_kwargs: Other kwargs for the node instance.

    Returns:
        (InputsToDict): A new node for transforming inputs into a dictionary.
    """
    cls = inputs_to_dict_factory(input_specification, class_name_suffix, use_cache)
    cls.preview_io()
    return cls(*node_args, **node_kwargs)


class InputsToDataframe(Transformer, ABC):
    """
    Turns inputs of dictionaries (all with the same keys) into a single
    :class:`pandas.DataFrame`.
    """

    _length: ClassVar[int]  # Mandatory attribute for non-abstract subclasses
    _output_name: ClassVar[str] = "df"

    @classmethod
    def _build_inputs_preview(cls) -> dict[str, tuple[Any, Any]]:
        return {f"row_{i}": (dict, NOT_DATA) for i in range(cls._length)}

    @classmethod
    def _build_outputs_preview(cls) -> dict[str, Any]:
        return {cls._output_name: DataFrame}

    @property
    def run_args(self) -> tuple[tuple, dict]:
        return tuple(self.inputs.to_value_dict().values()), {}

    def _on_run(self, *rows: dict[str, Any]) -> DataFrame:
        df_dict = {}
        for i, row in enumerate(rows):
            for key, value in row.items():
                if i == 0:
                    df_dict[key] = [value]
                else:
                    df_dict[key].append(value)
        return DataFrame(df_dict)

    def process_run_result(self, run_output: DataFrame) -> DataFrame:
        self.outputs[self._output_name].value = run_output
        return run_output


@classfactory
def inputs_to_dataframe_factory(
    n: int, use_cache: bool = True, /
) -> type[InputsToDataframe]:
    return (  # type: ignore[return-value]
        f"{InputsToDataframe.__name__}{n}",
        (InputsToDataframe,),
        {
            "_length": n,
            "use_cache": use_cache,
        },
        {},
    )


def inputs_to_dataframe(n: int, use_cache: bool = True, *node_args, **node_kwargs):
    """
    Creates and returns an instance of a dynamically generated
    :class:`InputsToDataframe` subclass with a specified number of inputs, each being a
    dictionary to form rows of the dataframe.

    Args:
        n (int): Number of input channels.
        use_cache (bool): Whether this node should default to caching its values.
            (Default is True.)
        *node_args: Positional arguments for the node instance.
        **node_kwargs: Keyword arguments for the node instance.

    Returns:
        InputsToDataframe: An instance of the dynamically created
            :class:`InputsToDataframe` subclass.
    """
    cls = inputs_to_dataframe_factory(n, use_cache)
    cls.preview_io()
    return cls(*node_args, **node_kwargs)


class DataclassNode(Transformer, ABC):
    """
    A base class for a node that converts inputs into a dataclass instance.
    """

    dataclass: ClassVar[type]  # Mandatory in children, must pass `is_dataclass`
    _output_name: ClassVar[str] = "dataclass"

    @classmethod
    def _dataclass_fields(cls):
        return cls.dataclass.__dataclass_fields__

    @classmethod
    def _build_inputs_preview(cls) -> dict[str, tuple[Any, Any]]:
        # Make a channel for each field
        return {
            name: (f.type, NOT_DATA if f.default is MISSING else f.default)
            for name, f in cls._dataclass_fields().items()
        }

    @classmethod
    def _build_outputs_preview(cls) -> dict[str, Any]:
        return {cls._output_name: cls.dataclass}

    def _setup_node(self) -> None:
        super()._setup_node()
        # Then leverage default factories from the dataclass
        for name, channel in self.inputs.items():
            if (
                channel.value is NOT_DATA
                and self._dataclass_fields()[name].default_factory is not MISSING
            ):
                self.inputs[name] = self._dataclass_fields()[name].default_factory()

    @property
    def run_args(self) -> tuple[tuple, dict]:
        return (), self.inputs.to_value_dict()

    def _on_run(self, **inputs_to_value_dict):
        return self.dataclass(**inputs_to_value_dict)

    def process_run_result(self, run_output: Any | tuple) -> Any | tuple:
        self.outputs[self._output_name].value = run_output
        return run_output

    @classmethod
    def _extra_info(cls) -> str:
        return cls.dataclass.__doc__ or ""


@classfactory
def dataclass_node_factory(
    dataclass: type, use_cache: bool = True, /
) -> type[DataclassNode]:
    if not isinstance(dataclass, type):
        raise TypeError(
            f"{DataclassNode} expected to get a dataclass but {dataclass} is not "
            f"type `type`."
        )
    dataclass = as_dataclass(dataclass)
    # Classes inheriting from a dataclass will pass the `dataclasses.is_dataclass` test
    # BUT they won't themselves _act_ as dataclass definitions! I.e. if you introduce
    # new fields in a sub-dataclass, or update defaults, this won't register _unless_
    # that new class is _also_ wrapped as a @dataclasses.dataclass!
    # This is not our fault, it's just a python thing. But is our _problem_.
    # To make sure dataclass nodes inheriting from other dataclass nodes still act as
    # dataclasses, just cast everything as a dataclass -- re-casting an
    # already-dataclass is not harmful
    # Composition is preferable over inheritance, but we want inheritance to be possible
    module, qualname = dataclass.__module__, dataclass.__qualname__
    dataclass.__qualname__ += ".dataclass"  # So output type hints know where to find it
    return (  # type: ignore[return-value]
        dataclass.__name__,
        (DataclassNode,),
        {
            "dataclass": dataclass,
            "__module__": module,
            "__qualname__": qualname,
            "__doc__": dataclass.__doc__,
            "use_cache": use_cache,
        },
        {},
    )


def as_dataclass_node(dataclass: type):
    """
    Decorates a dataclass as a dataclass node -- i.e. a node whose inputs correspond
    to dataclass fields and whose output is an instance of the dataclass.

    The underlying dataclass can be accessed on the :attr:`.dataclass` class attribute
    of the resulting node class.

    Leverages defaults (default factories) on dataclass fields to populate input
    channel values at class defintion (instantiation).

    Args:
        dataclass (type): A dataclass, i.e. class passing `dataclasses.is_dataclass`,
            or class definition that will be automatically wrapped with
            `dataclasses.dataclass`.
        use_cache (bool): Whether nodes of this type should default to caching their
            values. (Default is True.)

    Returns:
        (type[DataclassNode]): A :class:`DataclassNode` subclass whose instances
            transform inputs to an instance of that dataclass.

    Examples:
        >>> from dataclasses import dataclass, field
        >>>
        >>> from pyiron_workflow import Workflow
        >>>
        >>> def some_list():
        ...     return [1, 2, 3]
        >>>
        >>> @Workflow.wrap.as_dataclass_node
        ... class Foo:
        ...     necessary: str
        ...     bar: str = "bar"
        ...     answer: int = 42
        ...     complex_: list = field(default_factory=some_list)
        >>>
        >>> f = Foo()
        >>> print(f.readiness_report)
        Foo readiness report:
        ready: False
        running: False
        failed: False
        inputs.necessary: False
        inputs.bar: True
        inputs.answer: True
        inputs.complex_: True
        <BLANKLINE>

        >>> f(necessary="input as a node kwarg")
        Foo.dataclass(necessary='input as a node kwarg', bar='bar', answer=42, complex_=[1, 2, 3])
    """
    dataclass_node_factory.clear(dataclass.__name__)  # Force a fresh class
    module, qualname = dataclass.__module__, dataclass.__qualname__
    cls = dataclass_node_factory(dataclass)
    cls._reduce_imports_as = (module, qualname)
    cls.preview_io()
    return cls


def dataclass_node(dataclass: type, use_cache: bool = True, *node_args, **node_kwargs):
    """
    Builds a dataclass node from a dataclass -- i.e. a node whose inputs correspond
    to dataclass fields and whose output is an instance of the dataclass.

    The underlying dataclass can be accessed on the :attr:`.dataclass` class attribute
    of the resulting node.

    Leverages defaults (default factories) on dataclass fields to populate input
    channel values at class defintion (instantiation).

    Args:
        dataclass (type): A dataclass, i.e. class passing `dataclasses.is_dataclass`,
            or class variable that will be automatically passed to
            `dataclasses.dataclass`.
        use_cache (bool): Whether this node should default to caching its values.
            (Default is True.)
        *node_args: Other :class:`Node` positional arguments.
        **node_kwargs: Other :class:`Node` keyword arguments.

    Returns:
        (DataclassNode): An instance of the dynamically created :class:`DataclassNode`
            subclass.

    Examples:
        >>> from dataclasses import dataclass, field
        >>>
        >>> from pyiron_workflow import Workflow
        >>>
        >>> def some_list():
        ...     return [1, 2, 3]
        >>>
        >>> #@dataclass  # Works on actual dataclasses as well as dataclass-like classes
        >>> class Foo:
        ...     necessary: str
        ...     bar: str = "bar"
        ...     answer: int = 42
        ...     complex_: list = field(default_factory=some_list)
        >>>
        >>> f = Workflow.create.transformer.dataclass_node(Foo, label="my_dc")
        >>> print(f.readiness_report)
        my_dc readiness report:
        ready: False
        running: False
        failed: False
        inputs.necessary: False
        inputs.bar: True
        inputs.answer: True
        inputs.complex_: True
        <BLANKLINE>

        >>> f(necessary="input as a node kwarg")
        Foo.dataclass(necessary='input as a node kwarg', bar='bar', answer=42, complex_=[1, 2, 3])
    """
    cls = dataclass_node_factory(dataclass)
    cls.preview_io()
    return cls(*node_args, **node_kwargs)



"""
Classes for "lexical" reasoning.

The motivation here is to be able to provide the object with a unique identifier
in the context of other lexical objects. Each object may have at most one parent,
while lexical parents may have an arbitrary number of children, and each child's name
must be unique in the scope of that parent. In this way, when lexical parents are also
themselves lexical, we can build a path from the parent-most object to any child that
is completely unique. The typical filesystem on a computer is an excellent
example and fulfills our requirements, the only reason we depart from it is so that
we are free to have objects stored in different locations (possibly even on totally
different drives or machines) belong to the same lexical group.
"""

from __future__ import annotations

import contextlib
from abc import ABC, abstractmethod
from difflib import get_close_matches
from pathlib import Path
from typing import ClassVar, Generic, TypeVar

from bidict import bidict

from pyiron_workflow.logging import logger
from pyiron_workflow.mixin.has_interface_mixins import HasLabel, UsesState

ParentType = TypeVar("ParentType", bound="LexicalParent")


class Lexical(UsesState, HasLabel, Generic[ParentType], ABC):
    """
    An object with a unique lexical path.

    The lexical parent object (if any), and the parent-most object are both easily
    accessible.
    """

    lexical_delimiter: ClassVar[str] = "/"

    def __init__(
        self,
        *args,
        label: str | None = None,
        parent: ParentType | None = None,
        **kwargs,
    ):
        self._label = ""
        self._parent: ParentType | None = None
        self._detached_parent_path = None
        self.label = self.__class__.__name__ if label is None else label
        self.parent = parent
        super().__init__(*args, **kwargs)

    def _check_label(self, new_label: str) -> None:
        super()._check_label(new_label)
        if self.lexical_delimiter in new_label:
            raise ValueError(
                f"Lexical delimiter {self.lexical_delimiter} cannot be in new label "
                f"{new_label}"
            )

    @property
    def parent(self) -> ParentType | None:
        return self._parent

    @parent.setter
    def parent(self, new_parent: ParentType | None) -> None:
        self._set_parent(new_parent)

    def _set_parent(self, new_parent: ParentType | None):
        """
        mypy is uncooperative with super calls for setters, so we pull the behaviour
        out.
        """

        if new_parent is self._parent:
            # Exit early if nothing is changing
            return

        self._detached_parent_path = None

        if self.parent is not None and self in self.parent.children.values():
            old_parent = self.parent
            self._parent = None
            old_parent.remove_child(self)

        _ensure_path_is_not_cyclic(new_parent, self)
        self._parent = new_parent
        if new_parent is not None and self not in new_parent.children.values():
            new_parent.add_child(self)

    @property
    def lexical_path(self) -> str:
        """
        The path of node labels from the graph root (parent-most node) down to this
        node.
        """
        prefix: str
        if self.parent is None and self.detached_parent_path is None:
            prefix = ""
        elif self.parent is None and self.detached_parent_path is not None:
            prefix = self.detached_parent_path
        elif self.parent is not None and self.detached_parent_path is None:
            if isinstance(self.parent, Lexical):
                prefix = self.parent.lexical_path
            else:
                prefix = self.lexical_delimiter + self.parent.label
        else:
            raise ValueError(
                f"The parent and detached path should not be able to take non-None "
                f"values simultaneously, but got {self.parent} and "
                f"{self.detached_parent_path}, respectively. Please raise an issue on "
                f"GitHub outlining how your reached this state."
            )
        return prefix + self.lexical_delimiter + self.label

    @property
    def detached_parent_path(self) -> str | None:
        """
        The get/set state cycle of :class:`Lexical` de-parents objects, but we may
        still be interested in the lexical path -- e.g. if we `pickle` dump and load
        the object we will lose parent information, but this will still hold what the
        path _was_ before the orphaning process.

        The detached path will get cleared if a new parent is set, but is otherwise
        used as the root for the purposes of finding the lexical path.
        """
        return self._detached_parent_path

    @property
    def full_label(self) -> str:
        """
        A shortcut that combines the lexical path and label into a single string.
        """
        return self.lexical_path

    @property
    def lexical_root(self) -> Lexical:
        """The parent-most object in this lexical path; may be self."""
        if isinstance(self.parent, Lexical):
            return self.parent.lexical_root
        else:
            return self

    def as_path(self, root: Path | str | None = None, mkdir: bool = False) -> Path:
        """
        The lexical path as a :class:`pathlib.Path`, with a filesystem :param:`root`
        (default is the current working directory).
        """
        path = (Path.cwd() if root is None else Path(root)).joinpath(
            *self.lexical_path.split(self.lexical_delimiter)
        )
        if mkdir:
            path.mkdir(parents=True, exist_ok=True)
        return path

    def clean_path(
        self,
        root: Path | str | None = None,
        clean_parents: bool = True,
        remove_files: bool = False,
    ) -> None:
        """
        Recursively remove this object's directory and optionally its ancestors.
        Unless :param:`remove_files` is True, this will only remove empty directories.
        If non-lexical directories are present, they will always block removal.

        If this object is itself a lexical parent, it will run cleaning recursively on
        its children.

        Args:
            root: Base path from which to resolve the lexical path.
            clean_parents: If True, also attempt to remove parent directories.
            remove_files: If True, delete files within the directory before removal.
        """
        directory = self.as_path(root)

        if isinstance(self, LexicalParent):
            for child in self.children.values():
                child.clean_path(
                    root=root, clean_parents=False, remove_files=remove_files
                )

        if remove_files and directory.is_dir():
            for item in directory.iterdir():
                if item.is_file():
                    item.unlink()
        with contextlib.suppress(OSError):  # If it's not empty just move on
            directory.rmdir()

        if (
            clean_parents
            and self.parent is not None
            and isinstance(self.parent, Lexical)
        ):
            self.parent.clean_path(
                root=root, clean_parents=clean_parents, remove_files=remove_files
            )

    def __getstate__(self):
        state = super().__getstate__()
        if self.parent is not None:
            state["_detached_parent_path"] = self.parent.lexical_path
        state["_parent"] = None
        # Regarding removing parent from state:
        # Basically we want to avoid recursion during (de)serialization; when the
        # parent object is deserializing itself, _it_ should know who its children are
        # and inform them of this.
        # In the case the object gets passed to another process using __getstate__,
        # this also avoids dragging our whole lexical parent graph along with us.
        return state


class CyclicPathError(ValueError):
    """
    To be raised when adding a child would result in a cyclic lexical path.
    """


class AlreadyHasParentError(ValueError):
    """
    To be raised when parenting an already-parented child.
    """


ChildType = TypeVar("ChildType", bound=Lexical)


class LexicalParent(HasLabel, Generic[ChildType], ABC):
    """
    A labeled object with a collection of uniquely-named lexical children.

    Children should be added or removed via the :meth:`add_child` and
    :meth:`remove_child` methods and _not_ by direct manipulation of the
    :attr:`children` container.

    Children are dot-accessible and appear in :meth:`__dir__` for tab-completion.

    Iterating over the parent yields the children, and the length of the parent is
    the number of children.

    When adding children or assigning parents, a check is performed on the lexical
    path to forbid cyclic paths.
    """

    def __init__(
        self,
        *args,
        strict_naming: bool = True,
        **kwargs,
    ):
        self._children: bidict[str, ChildType] = bidict()
        self.strict_naming = strict_naming
        super().__init__(*args, **kwargs)

    @classmethod
    @abstractmethod
    def child_type(cls) -> type[ChildType]:
        # Dev note: In principle, this could be a regular attribute
        # However, in other situations this is precluded (e.g. in channels)
        # since it would result in circular references.
        # Here we favour consistency over brevity,
        # and maintain the X_type() class method pattern
        pass

    @property
    def children(self) -> bidict[str, ChildType]:
        return self._children

    @property
    def child_labels(self) -> tuple[str]:
        return tuple(child.label for child in self)

    def _check_label(self, new_label: str) -> None:
        super()._check_label(new_label)
        if self.child_type().lexical_delimiter in new_label:
            raise ValueError(
                f"Child type ({self.child_type()}) lexical delimiter "
                f"{self.child_type().lexical_delimiter} cannot be in new label "
                f"{new_label}"
            )

    def __getattr__(self, key) -> ChildType:
        try:
            return self._children[key]
        except KeyError as key_error:
            # Raise an attribute error from getattr to make sure hasattr works well!
            msg = f"Could not find attribute '{key}' on {self.label} "
            msg += f"({self.__class__.__name__}) or among its children "
            msg += f"({self._children.keys()})."
            matches = get_close_matches(key, self._children.keys(), cutoff=0.8)
            if len(matches) > 0:
                msg += f" Did you mean '{matches[0]}' and not '{key}'?"
            raise AttributeError(msg) from key_error

    def __iter__(self):
        return self.children.values().__iter__()

    def __len__(self) -> int:
        return len(self.children)

    def __dir__(self):
        return set(super().__dir__() + list(self.children.keys()))

    def add_child(
        self,
        child: ChildType,
        label: str | None = None,
        strict_naming: bool | None = None,
    ) -> ChildType:
        """
        Add a child, optionally assigning it a new label in the process.

        Args:
            child (ChildType): The child to add.
            label (str|None): A (potentially) new label to assign the child. (Default
                is None, leave the child's label alone.)
            strict_naming (bool|None): Whether to append a suffix to the label if
                another child is already held with the same label. (Default is None,
                use the class-level flag.)

        Returns:
            (ChildType): The child being added.

        Raises:
            TypeError: When the child is not of an allowed class.
            ValueError: When the child has a different parent already.
            AttributeError: When the label is already an attribute (but not a child).
            AttributeError: When the label conflicts with another child and
                `strict_naming` is true.

        """
        if not isinstance(child, self.child_type()):
            raise TypeError(
                f"{self.label} expected a new child of type {self.child_type()} "
                f"but got {child}"
            )

        self._ensure_child_has_no_other_parent(child)

        _ensure_path_is_not_cyclic(self, child)

        label = child.label if label is None else label
        strict_naming = self.strict_naming if strict_naming is None else strict_naming

        if self._this_child_is_already_at_this_label(child, label):
            pass
        else:
            label = self._get_unique_label(label, strict_naming)

            if self._this_child_is_already_at_a_different_label(child, label):
                self.children.inv.pop(child)

            # Finally, update label and reflexively form the parent-child relationship
            child.label = label
            self.children[child.label] = child
            child.parent = self
        return child

    def _ensure_child_has_no_other_parent(self, child: Lexical) -> None:
        if child.parent is not None and child.parent is not self:
            raise AlreadyHasParentError(
                f"The child ({child.label}) already belongs to the parent "
                f"{child.parent.label}. Please remove it there before trying to "
                f"add it to this parent ({self.label})."
            )

    def _this_child_is_already_at_this_label(self, child: Lexical, label: str) -> bool:
        return (
            label == child.label
            and label in self.child_labels
            and self.children[label] is child
        )

    def _this_child_is_already_at_a_different_label(self, child, label) -> bool:
        return child.parent is self and label != child.label

    def _get_unique_label(self, label: str, strict_naming: bool) -> str:
        if label in self.__dir__():
            if label in self.child_labels:
                if strict_naming:
                    raise AttributeError(
                        f"{label} is already the label for a child. Please remove it "
                        f"before assigning another child to this label."
                    )
                else:
                    label = self._add_suffix_to_label(label)
            else:
                raise AttributeError(
                    f"{label} is an attribute or method of the {self.__class__} class, "
                    f"and cannot be used as a child label."
                )
        return label

    def _add_suffix_to_label(self, label: str) -> str:
        i = 0
        new_label = label
        while new_label in self.__dir__():
            # We search dir and not just the child_labels for the edge case that
            # someone has a very label-like attribute
            new_label = f"{label}{i}"
            i += 1
        if new_label != label:
            logger.info(
                f"{label} is already a node; appending an index to the "
                f"node label instead: {new_label}"
            )
        return new_label

    def remove_child(self, child: ChildType | str) -> ChildType:
        if isinstance(child, str):
            child_instance = self.children.pop(child)
        elif isinstance(child, self.child_type()):
            self.children.inv.pop(child)
            child_instance = child
        else:
            raise TypeError(
                f"{self.label} expected to remove a child of type str or "
                f"{self.child_type()} but got {child}"
            )

        child_instance.parent = None

        return child_instance

    def __getstate__(self):
        state = super().__getstate__()

        # Remove the children from the state and store each element right in the state
        # -- the labels are guaranteed to not be attributes already so this is safe,
        # and it makes sure that the state path matches the lexical path
        del state["_children"]
        state["child_labels"] = self.child_labels
        for child in self:
            state[child.label] = child

        return state

    def __setstate__(self, state):
        # Reconstruct children from state
        # Remove them from the state as you go, so they don't hang around in the
        # __dict__ after we set state -- they were only there to start with to guarantee
        # that the state path and the lexical path matched (i.e. without ".children."
        # in between)
        state["_children"] = bidict(
            {label: state.pop(label) for label in state.pop("child_labels")}
        )

        super().__setstate__(state)

        self._children = bidict(self._children)

        # Children purge their parent information in their __getstate__. This avoids
        # recursion, so we don't need to ship an entire graph off to a second process,
        # but rather can send just the requested object and its scope (lexical
        # children). So, now return their parent to them:
        for child in self:
            child.parent = self


def _ensure_path_is_not_cyclic(parent, child: Lexical) -> None:
    if isinstance(parent, Lexical) and parent.lexical_path.startswith(
        child.lexical_path + child.lexical_delimiter
    ):
        raise CyclicPathError(
            f"{parent.label} cannot be the parent of {child.label}, because its "
            f"lexical path is already in {child.label}'s path and cyclic paths "
            f"are not allowed. (i.e. {child.lexical_path} is in "
            f"{parent.lexical_path})"
        )



"""
Collections of channel objects.

These also support the syntactic sugar of treating value assignments and new
connections on the same footing.
"""

from __future__ import annotations

import contextlib
from abc import ABC, abstractmethod
from collections.abc import ItemsView, Iterator
from typing import Any, Generic, TypeVar

from pyiron_snippets.dotdict import DotDict

from pyiron_workflow.channels import (
    Channel,
    DataChannel,
    InputData,
    InputSignal,
    InputType,
    OutputData,
    OutputSignal,
    OutputType,
    SignalChannel,
)
from pyiron_workflow.logging import logger
from pyiron_workflow.mixin.display_state import HasStateDisplay
from pyiron_workflow.mixin.has_interface_mixins import HasChannel

OwnedType = TypeVar("OwnedType", bound=Channel)
OwnedConjugate = TypeVar("OwnedConjugate", bound=Channel)


class IO(HasStateDisplay, Generic[OwnedType, OwnedConjugate], ABC):
    """
    IO is a convenience layer for holding and accessing multiple input/output channels.
    It allows key and dot-based access to the underlying channels.
    Channels can also be iterated over, and there are a number of helper functions to
    alter the properties of or check the status of all the channels at once.

    A new channel can be assigned as an attribute of an IO collection, as long as it
    matches the channel's type (e.g. :class:`OutputChannel` for :class:`Outputs`, :class:`InputChannel`
    for :class:`Inputs`, etc...).

    When assigning something to an attribute holding an existing channel, if the
    assigned object is a :class:`Channel`, then an attempt is made to make a :attr:`connection`
    between the two channels, otherwise we fall back on a value assignment that must
    be defined in child classes under `_assign_value_to_existing_channel`.
    This provides syntactic sugar such that both new connections and new values can
    be assigned with a simple `=`.
    """

    channel_dict: DotDict[str, OwnedType]

    def __init__(self, *channels: OwnedType) -> None:
        self.__dict__["channel_dict"] = DotDict(
            {
                channel.label: channel
                for channel in channels
                if isinstance(channel, self._channel_class)
            }
        )

    @property
    @abstractmethod
    def _channel_class(self) -> type[OwnedType]:
        pass

    @abstractmethod
    def _assign_a_non_channel_value(self, channel: OwnedType, value: Any) -> None:
        """What to do when some non-channel value gets assigned to a channel"""
        pass

    def __getattr__(self, item: str) -> OwnedType:
        try:
            return self.channel_dict[item]
        except KeyError as key_error:
            # Raise an attribute error from getattr to make sure hasattr works well!
            raise AttributeError(
                f"Could not find attribute {item} on {self.__class__.__name__} object "
                f"nor in its channels ({self.labels})"
            ) from key_error

    def __setattr__(self, key: str, value: Any) -> None:
        if key in self.channel_dict:
            self._assign_value_to_existing_channel(self.channel_dict[key], value)
        elif isinstance(value, self._channel_class):
            if key != value.label:
                logger.info(
                    f"Assigning a channel with the label {value.label} to the io key "
                    f"{key}"
                )
            self.channel_dict[key] = value
        else:
            raise TypeError(
                f"Can only set Channel object or connect to existing channels, but the "
                f"attribute {key} got assigned {value} of type {type(value)}"
            )

    def _assign_value_to_existing_channel(self, channel: OwnedType, value: Any) -> None:
        if isinstance(value, HasChannel):
            self._assign_a_channel_value(channel, value)
        else:
            self._assign_a_non_channel_value(channel, value)

    def _assign_a_channel_value(self, channel: OwnedType, value: HasChannel) -> None:
        channel.connect(value.channel)

    def __getitem__(self, item: str) -> OwnedType:
        return self.__getattr__(item)

    def __setitem__(self, key: str, value: Any) -> None:
        self.__setattr__(key, value)

    @property
    def connections(self) -> list[OwnedConjugate]:
        """All the unique connections across all channels"""
        return list(
            {connection for channel in self for connection in channel.connections}
        )

    @property
    def connected(self) -> bool:
        return any(c.connected for c in self)

    @property
    def fully_connected(self) -> bool:
        return all(c.connected for c in self)

    def disconnect(self) -> list[tuple[OwnedType, OwnedConjugate]]:
        """
        Disconnect all connections that owned channels have.

        Returns:
            [list[tuple[Channel, Channel]]]: A list of the pairs of channels that no
                longer participate in a connection.
        """
        destroyed_connections = []
        for c in self:
            destroyed_connections.extend(c.disconnect_all())
        return destroyed_connections

    @property
    def labels(self) -> list[str]:
        return list(self.channel_dict.keys())

    def items(self) -> ItemsView[str, OwnedType]:
        return self.channel_dict.items()

    def __iter__(self) -> Iterator[OwnedType]:
        return self.channel_dict.values().__iter__()

    def __len__(self) -> int:
        return len(self.channel_dict)

    def __dir__(self):
        return list(set(super().__dir__() + self.labels))

    def __str__(self) -> str:
        return f"{self.__class__.__name__} {self.labels}"

    def __getstate__(self) -> dict[str, Any]:
        # Compatibility with python <3.11
        return dict(self.__dict__)

    def __setstate__(self, state: dict[str, Any]) -> None:
        # Because we override getattr, we need to use __dict__ assignment directly in
        # __setstate__ the same way we need it in __init__
        self.__dict__["channel_dict"] = state["channel_dict"]

    def display_state(
        self, state: dict[str, Any] | None = None, ignore_private: bool = True
    ) -> dict[str, Any]:
        state = dict(self.__getstate__()) if state is None else state
        for k, v in state["channel_dict"].items():
            state[k] = v
        del state["channel_dict"]
        return super().display_state(state=state, ignore_private=ignore_private)


class InputsIO(IO[InputType, OutputType], ABC):
    pass


class OutputsIO(IO[OutputType, InputType], ABC):
    pass


class DataIO(IO[DataChannel, DataChannel], ABC):
    def _assign_a_non_channel_value(self, channel: DataChannel, value) -> None:
        channel.value = value

    def to_value_dict(self) -> dict[str, Any]:
        return {label: channel.value for label, channel in self.channel_dict.items()}

    def to_list(self) -> list[Any]:
        """A list of channel values (order not guaranteed)"""
        return [channel.value for channel in self.channel_dict.values()]

    @property
    def ready(self) -> bool:
        return all(c.ready for c in self)

    def activate_strict_hints(self):
        [c.activate_strict_hints() for c in self]

    def deactivate_strict_hints(self):
        [c.deactivate_strict_hints() for c in self]


class Inputs(InputsIO, DataIO):
    @property
    def _channel_class(self) -> type[InputData]:
        return InputData

    def fetch(self) -> None:
        for c in self:
            c.fetch()

    def _assign_a_channel_value(self, channel: OwnedType, value: Any) -> None:
        # Allow the owned input data channel to overwrite its connection
        channel.disconnect_all()
        super()._assign_a_channel_value(channel, value)


OutputDataType = TypeVar("OutputDataType", bound=OutputData)


class GenericOutputs(OutputsIO, DataIO, Generic[OutputDataType], ABC):
    @property
    @abstractmethod
    def _channel_class(self) -> type[OutputDataType]:
        pass


class Outputs(GenericOutputs[OutputData]):
    @property
    def _channel_class(self) -> type[OutputData]:
        return OutputData


class SignalIO(IO[SignalChannel, SignalChannel], ABC):
    def _assign_a_non_channel_value(self, channel: SignalChannel, value: Any) -> None:
        raise TypeError(
            f"Tried to assign {value} ({type(value)} to the {channel.full_label}, "
            f"which is already a {type(channel)}. Only other signal channels may be "
            f"connected in this way."
        )


class InputSignals(InputsIO, SignalIO):
    @property
    def _channel_class(self) -> type[InputSignal]:
        return InputSignal

    def disconnect_run(self) -> list[tuple[InputSignal, OutputSignal]]:
        """Disconnect all `run` and `accumulate_and_run` signals, if they exist."""
        disconnected = []
        with contextlib.suppress(AttributeError):
            disconnected += self.run.disconnect_all()
        with contextlib.suppress(AttributeError):
            disconnected += self.accumulate_and_run.disconnect_all()
        return disconnected


class OutputSignals(OutputsIO, SignalIO):
    @property
    def _channel_class(self) -> type[OutputSignal]:
        return OutputSignal


class Signals(HasStateDisplay):
    """
    A meta-container for input and output signal IO containers.

    Attributes:
        input (InputSignals): An empty input signals IO container.
        output (OutputSignals): An empty input signals IO container.
    """

    def __init__(self) -> None:
        self.input: InputSignals = InputSignals()
        self.output: OutputSignals = OutputSignals()

    def disconnect(self) -> list[tuple[SignalChannel, SignalChannel]]:
        """
        Disconnect all connections in input and output signals.

        Returns:
            [list[tuple[Channel, Channel]]]: A list of the pairs of channels that no
                longer participate in a connection.
        """
        return self.input.disconnect() + self.output.disconnect()

    def disconnect_run(self) -> list[tuple[InputSignal, OutputSignal]]:
        return self.input.disconnect_run()

    @property
    def connected(self) -> bool:
        return self.input.connected or self.output.connected

    @property
    def fully_connected(self) -> bool:
        return self.input.fully_connected and self.output.fully_connected

    def __str__(self) -> str:
        return f"{str(self.input)}\n{str(self.output)}"



import keyword
import re


def to_identifier(s: str) -> str:
    # Replace invalid chars and ensure it doesn't start with a digit
    s = re.sub(r"\W|^(?=\d)", "_", s)
    # Append underscore if it's a keyword
    if keyword.iskeyword(s):
        s += "_"
    return s



"""
Channels are access points for information to flow into and out of nodes.
They accomplish this by forming connections between each other, and it should be as
easy as possible to form sensible and reliable connections.

Nodes get the attention, but channels are the real heroes.
"""

from __future__ import annotations

import inspect
import typing
from abc import ABC, abstractmethod

from pyiron_workflow.data import NOT_DATA
from pyiron_workflow.mixin.display_state import HasStateDisplay
from pyiron_workflow.mixin.has_interface_mixins import HasChannel, HasLabel
from pyiron_workflow.type_hinting import (
    type_hint_is_as_or_more_specific_than,
    valid_value,
)

if typing.TYPE_CHECKING:
    from pyiron_workflow.node import Node


class ChannelError(Exception):
    pass


class ChannelConnectionError(ChannelError):
    pass


class TooManyConnectionsError(ChannelError):
    pass


class InputLockedError(ChannelError, ValueError): ...


ConjugateType = typing.TypeVar("ConjugateType", bound="Channel")
InputType = typing.TypeVar("InputType", bound="InputChannel")
OutputType = typing.TypeVar("OutputType", bound="OutputChannel")
FlavorType = typing.TypeVar("FlavorType", bound="FlavorChannel")
ReceiverType = typing.TypeVar("ReceiverType", bound="DataChannel")


class Callback:
    def __init__(
        self,
        callback: typing.Callable[..., typing.Any] | None,
        *args: typing.Any,
        **kwargs: typing.Any,
    ):
        self.callback = callback
        self.args = args
        self.kwargs = kwargs

    def call(self):
        if self.callback is not None:
            self.callback(*self.args, **self.kwargs)


class Channel(
    HasChannel, HasLabel, HasStateDisplay, typing.Generic[ConjugateType], ABC
):
    """
    Channels facilitate the flow of information (data or control signals) into and
    out of :class:`Node` objects (namely nodes).

    They must have an identifier (`label: str`) and belong to an
    `owner: pyiron_workflow.node.Node`.


    Channels may form (:meth:`connect`/:meth:`disconnect`) and store
    (:attr:`connections`) connections with other channels.

    This connection information is reflexive, and is duplicated to be stored on _both_
    channels in the form of a reference to their counterpart in the connection.

    By using the provided methods to modify connections, the reflexive nature of
    these (dis)connections is guaranteed to be handled, and new connections are
    subjected to a validity test.

    Child classes must specify a conjugate class in order to enforce connection
    conjugate pairs which have the same "flavor" (e.g. "data" or "signal"), and
    opposite "direction" ("input" vs "output"). And they must define a string
    representation.

    Iterating over channels yields their connections.

    The length of a channel is the length of its connections.

    Attributes:
        owner (pyiron_workflow.node.Node): The channel's owner.
        connections (list[Channel]): Other channels to which this channel is connected.
    """

    def __init__(
        self,
        label: str,
        owner: Node,
    ):
        """
        Make a new channel.

        Args:
            label (str): A name for the channel.
            owner (pyiron_workflow.node.Node): The channel's owner.
        """
        self._label = label
        self.owner: Node = owner
        self.connections: list[ConjugateType] = []

    @abstractmethod
    def __str__(self):
        pass

    @classmethod
    @abstractmethod
    def connection_conjugate(cls) -> type[ConjugateType]:
        """
        The class forming a conjugate pair with this channel class -- i.e. the same
        "flavor" of channel, but opposite in I/O.
        """

    @property
    def scoped_label(self) -> str:
        """A label combining the channel's usual label and its owner's label"""
        return f"{self.owner.label}__{self.label}"

    @property
    def full_label(self) -> str:
        """A label combining the channel's usual label and its owner's lexical path"""
        return f"{self.owner.full_label}.{self.label}"

    def connect(self, *others: ConjugateType) -> None:
        """
        Form a connection between this and one or more other channels.
        Connections are reflexive, and should only occur between input and output
        channels, i.e. they are instances of each others
        :meth:`connection_conjugate()`.

        New connections get _prepended_ to the connection lists, so they appear first
        when searching over connections.

        Args:
            *others (Channel): The other channel objects to attempt to connect with.

        Raises:
            (ChannelConnectionError): If the other channel is of the correct type, but
                nonetheless not a valid connection.
            (TypeError): If the other channel is not an instance of this channel's
                partner type.
        """
        for other in others:
            if other in self.connections:
                continue
            elif isinstance(other, self.connection_conjugate()):
                wrong_parents_callback = self._ensure_same_owner_parent(other)
                if self._valid_connection(other) and other._valid_connection(self):
                    # Prepend new connections
                    # so that connection searches run newest to oldest
                    self.connections.insert(0, other)
                    other.connections.insert(0, self)
                else:
                    wrong_parents_callback.call()
                    raise ChannelConnectionError(
                        self._connection_conjugate_failure_message(other)
                    ) from None
            else:
                raise TypeError(
                    f"Can only connect to {self.connection_conjugate()} "
                    f"objects, but {self.full_label} ({self.__class__}) "
                    f"got {other} ({type(other)})"
                )

    def _ensure_same_owner_parent(self, other: ConjugateType) -> Callback:
        """
        We only want to form connections between channels in the same graph, but we
        so make sure that the channels' owners have the same parent, or fail.
        """
        if self.owner.parent is not None and other.owner.parent is None:
            self.owner.parent.add_child(other.owner)
            return Callback(self.owner.parent.remove_child, other.owner)
        elif self.owner.parent is None and other.owner.parent is not None:
            other.owner.parent.add_child(self.owner)
            return Callback(other.owner.parent.remove_child, self.owner)
        elif (
            self.owner.parent is not None
            and other.owner.parent is not None
            and self.owner.parent is not other.owner.parent
        ):
            raise ChannelConnectionError(
                f"Can only connect channels inside the same graph, but "
                f"{self.full_label} has the owner {self.owner.full_label} with the "
                f"parent {self.owner.parent.full_label} and {other.full_label} has "
                f"the owner {other.owner.full_label} with the parent "
                f"{other.owner.parent.full_label}."
            )
        return Callback(None)

    def _valid_connection(self, other: ConjugateType) -> bool:
        """
        Logic for determining if a connection to a conjugate partner is valid.

        Override in child classes as necessary.
        """
        return True

    def _connection_conjugate_failure_message(self, other: ConjugateType) -> str:
        return (
            f"The channel {other.full_label} ({other.__class__}) has the "
            f"correct type ({self.connection_conjugate()}) to connect with "
            f"{self.full_label} ({self.__class__}), but is not a valid "
            f"connection."
        )

    def disconnect(
        self, *others: ConjugateType
    ) -> list[tuple[typing.Self, ConjugateType]]:
        """
        If currently connected to any others, removes this and the other from eachothers
        respective connections lists.

        Args:
            *others (Channel): The other channels to disconnect from.

        Returns:
            [list[tuple[Channel, Channel]]]: A list of the (input, output) conjugate
                pairs of channels that no longer participate in a connection.
        """
        destroyed_connections = []
        for other in others:
            if other in self.connections:
                self.connections.remove(other)
                other.disconnect(self)
                destroyed_connections.append((self, other))
        return destroyed_connections

    def disconnect_all(self) -> list[tuple[typing.Self, ConjugateType]]:
        """
        Disconnect from all other channels currently in the connections list.
        """
        return self.disconnect(*self.connections)

    @property
    def connected(self) -> bool:
        """
        Has at least one connection.
        """
        return len(self.connections) > 0

    def __iter__(self):
        return self.connections.__iter__()

    @property
    def channel(self) -> typing.Self:
        return self

    def move_connections(self, other: typing.Self) -> None:
        """
        Adds all the connections in another channel to this channel's connections.

        If an exception is encountered, all the new connections are disconnected before
        the exception is raised.
        """
        original_partners = [c[1] for c in self.disconnect_all()]
        other_partners = [c[1] for c in other.disconnect_all()]
        try:
            self.connect(*other_partners)
        except Exception as e:
            self.disconnect_all()
            self.connect(*original_partners)
            other.connect(*other_partners)
            raise e

    def __getstate__(self):
        state = super().__getstate__()
        state["connections"] = []
        # It is the responsibility of the owner's parent to store and restore
        # connections (if any), since these can extend beyond the owner and would thus
        # bloat the data being sent cross-process if the owner is shipped off
        return state

    def display_state(self, state=None, ignore_private=True):
        state = dict(self.__getstate__()) if state is None else state
        state["owner"] = state["owner"].full_label  # JSON can't handle recursion
        state["connections"] = [c.full_label for c in self.connections]
        return super().display_state(state=state, ignore_private=ignore_private)


class FlavorChannel(Channel[FlavorType], ABC):
    """Abstract base for all flavor-specific channels."""


class InputChannel(Channel[OutputType], ABC):
    """Mixin for input channels."""


class OutputChannel(Channel[InputType], ABC):
    """Mixin for output channels."""


class DataChannel(FlavorChannel["DataChannel"], typing.Generic[ReceiverType], ABC):
    """
    Data channels control the flow of data on the graph.

    They store data persistently (:attr:`value`).

    This value may have a default (:attr:`default`) and the default-default is to be
    `NOT_DATA`.

    They may optionally have a type hint (:attr:`type_hint`).

    New data and new connections are tested against type hints (if any).

    In addition to the requirement of being a "conjugate pair", if both connecting
    channels have type hints, the output channel must have a type hint that is as or
    more specific than the input channel.

    In addition to connections, these channels can have a single partner
    (:attr:`value_receiver`) that is of the same data flavor and the same direction
    (i.e. input or output) and obeys type hints as though it were the "downstream"
    (input) partner in a connection.
    Channels with such partners pass any data updates they receive directly to this
    partner (via the :attr:`value` setter).
    (This is helpful for passing data between scopes, where we want input at one scope
    to be passed to the input of owners at a deeper scope, i.e. macro input passing to
    child node input, or vice versa for output.)

    All these type hint tests can be disabled on the input/receiving channel
    (:attr:`strict_hints: bool`), and this is recommended for the optimal performance
    in production runs.

    Channels can indicate whether they hold data they are happy with
    (:attr:`ready: bool`), which is to say it is data (not the singleton `NOT_DATA`)
    and that it conforms to the type hint (if one is provided and checking is active).

    TODO:

        - Storage (including priority and history)
        - Ontological hinting

    Some comments on type hinting:
    For simple type hints like `int` or `str`, type hint comparison is trivial.
    However, some hints take arguments, e.g. `dict[str, int]` to specify key and value
    types; `tuple[int, int, str]` to specify a tuple with certain values;
    `typing.Literal['a', 'b', 'c']` to specify particular choices;
    `typing.Callable[[float, float], str]` to specify a callable that takes particular
    argument types and has a return type; etc.
    For hints with the origin `dict`, `tuple`, and `typing.Callable`, the two hints must
    have _exactly the same arguments_ for one two qualify as "as or more specific".
    E.g. `tuple[int, int|float]` is as or more specific than
    `tuple[int|float, int|float]`, but not `tuple[int, int|float, str]`.
    For _all other hints_, we demand that the output hint arguments be a _subset_ of
    the input.
    E.g. `Literal[1, 2]` is as or more specific that both `Literal[1, 2]` and
    `Literal[1, 2, "three"]`.

    Warning:
        Type hinting in python is quite complex, and determining when a hint is
        "more specific" can be tricky. For instance, in python 3.11 you can now type
        hint a tuple with a mixture of fixed elements of fixed type, followed by an
        arbitrary elements of arbitrary type. This and other complex scenarios are not
        yet included in our test suite and behaviour is not guaranteed.

    Attributes:
        value: The actual data value held by the channel.
        owner (pyiron_workflow.node.Node): The channel's owner.
        default (typing.Any|None): The default value to initialize to.
            (Default is the singleton `NOT_DATA`.)
        type_hint (typing.Any|None): A type hint for values. (Default is None.)
        strict_hints (bool): Whether to check new values, connections, and partners
            when this channel is a value receiver. This can potentially be expensive, so
            consider deactivating strict hints everywhere for production runs. (Default
            is True, raise exceptions when type hints get violated.)
        value_receiver (ReceiverType|None): Another channel of the receiver type (i.e.
            also a data flavor and matching input/output type) whose value will always
            get updated when this channel's value gets updated.
    """

    def __init__(
        self,
        label: str,
        owner: Node,
        default: typing.Any | None = NOT_DATA,
        type_hint: typing.Any | None = None,
        strict_hints: bool = True,
        value_receiver: ReceiverType | None = None,
    ):
        super().__init__(label=label, owner=owner)
        self._value = NOT_DATA
        self._value_receiver: ReceiverType | None = None
        self.type_hint = type_hint
        self.strict_hints = strict_hints
        self.default = default
        self.value = default  # Implicitly type check your default by assignment
        self.value_receiver = value_receiver

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        self._type_check_new_value(new_value)
        if self.value_receiver is not None:
            self.value_receiver.value = new_value
        self._value = new_value

    def _type_check_new_value(self, new_value):
        if (
            self.strict_hints
            and new_value is not NOT_DATA
            and self._has_hint
            and not valid_value(new_value, self.type_hint)
        ):
            raise TypeError(
                f"The channel {self.full_label} cannot take the value `{new_value}` "
                f"({type(new_value)}) because it is not compliant with the type hint "
                f"{self.type_hint}"
            )

    @property
    def value_receiver(self) -> ReceiverType | None:
        """
        Another data channel of the same type to whom new values are always pushed
        (without type checking of any sort, not even when forming the couple!)

        Useful for macros, so that the IO of children and IO at the macro level can
        be kept synchronized.
        """
        return self._value_receiver

    @value_receiver.setter
    def value_receiver(self, new_partner: ReceiverType | None):
        if new_partner is not None:
            if not isinstance(new_partner, self.__class__):
                raise TypeError(
                    f"The {self.__class__.__name__} {self.full_label} got a coupling "
                    f"partner {new_partner} but requires something of the same type"
                )

            if new_partner is self:
                raise ValueError(
                    f"{self.__class__.__name__} {self.full_label} cannot couple to "
                    f"itself"
                )

            if (
                self._both_typed(new_partner)
                and new_partner.strict_hints
                and not type_hint_is_as_or_more_specific_than(
                    self.type_hint, new_partner.type_hint
                )
            ):
                raise ValueError(
                    f"The channel {self.full_label} cannot take "
                    f"{new_partner.full_label} as a value receiver because this "
                    f"type hint ({self.type_hint}) is not as or more specific than "
                    f"the receiving type hint ({new_partner.type_hint})."
                )

            new_partner.value = self.value

        self._value_receiver = typing.cast(ReceiverType, new_partner)

    @property
    def ready(self) -> bool:
        """
        Check if the currently stored value is data and satisfies the channel's type
        hint (if hint checking is activated).

        Returns:
            (bool): Whether the value is data and matches the type hint.
        """
        return self._value_is_data and (
            valid_value(self.value, self.type_hint)
            if self._has_hint and self.strict_hints
            else True
        )

    @property
    def _value_is_data(self) -> bool:
        return self.value is not NOT_DATA

    @property
    def _has_hint(self) -> bool:
        return self.type_hint is not None

    def _valid_connection(self, other: DataChannel) -> bool:
        if self._both_typed(other):
            out, inp = self._figure_out_who_is_who(other)
            if not inp.strict_hints:
                return True
            else:
                return type_hint_is_as_or_more_specific_than(
                    out.type_hint, inp.type_hint
                )
        else:
            # If either is untyped, don't do type checking
            return True

    def _connection_conjugate_failure_message(self, other: DataChannel) -> str:
        msg = super()._connection_conjugate_failure_message(other)
        msg += (
            f"Please check type hints, etc. {other.full_label}.type_hint = "
            f"{other.type_hint}; {self.full_label}.type_hint = {self.type_hint}"
        )
        return msg

    def _both_typed(self, other: DataChannel) -> bool:
        return self._has_hint and other._has_hint

    def _figure_out_who_is_who(
        self, other: DataChannel
    ) -> tuple[OutputData, InputData]:
        if isinstance(self, InputData) and isinstance(other, OutputData):
            return other, self
        elif isinstance(self, OutputData) and isinstance(other, InputData):
            return self, other
        else:
            raise ChannelError(
                f"This should be unreachable; data channel conjugate pairs should "
                f"always be input/output, but got {type(self)} for {self.full_label} "
                f"and {type(other)} for {other.full_label}. If you don't believe you "
                f"are responsible for this error, please contact the maintainers via "
                f"GitHub."
            )

    def __str__(self):
        return str(self.value)

    def activate_strict_hints(self) -> None:
        self.strict_hints = True

    def deactivate_strict_hints(self) -> None:
        self.strict_hints = False

    def __getstate__(self):
        state = super().__getstate__()
        state["_value_receiver"] = None
        # Value receivers live in the scope of Macros, so (re)storing them is the
        # owning macro's responsibility
        return state

    def display_state(self, state=None, ignore_private=True):
        state = dict(self.__getstate__()) if state is None else state
        self._make_entry_public(state, "_value", "value")
        return super().display_state(state=state, ignore_private=ignore_private)


class InputData(DataChannel["InputData"], InputChannel["OutputData"]):

    @classmethod
    def connection_conjugate(cls) -> type[OutputData]:
        return OutputData

    def _valid_connection(self, other: DataChannel[typing.Any]) -> bool:
        if len(self.connections) > 0:
            raise TooManyConnectionsError(
                f"{self.full_label} is already connected to "
                f"{self.connections[0].full_label} -- disconnect first before trying "
                f"to connect to {other.full_label}"
            )
        return super()._valid_connection(other)

    def fetch(self) -> None:
        """
        Sets :attr:`value` to the first value among connections (i.e. the most recent)
        that is something other than `NOT_DATA`; if no such value exists (e.g. because
        there are no connections or because all the connected output channels have
        `NOT_DATA` as their value), :attr:`value` remains unchanged.
        I.e., the connection with the highest priority for updating input data is the
        0th connection; build graphs accordingly.

        Raises:
            RuntimeError: If the owner is :attr:`running`.
        """
        for out in self.connections:
            if out.value is not NOT_DATA:
                self.value = out.value
                break

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        if self.owner.data_input_locked():
            raise InputLockedError(
                f"Owner {self.full_label} has its data input locked, "
                f"so value cannot be updated."
            )
        self._type_check_new_value(new_value)
        if self.value_receiver is not None:
            self.value_receiver.value = new_value
        self._value = new_value


class OutputData(DataChannel["OutputData"], OutputChannel["InputData"]):
    @classmethod
    def connection_conjugate(cls) -> type[InputData]:
        return InputData


SignalType = typing.TypeVar("SignalType", bound="SignalChannel")


class SignalChannel(FlavorChannel[SignalType], ABC):
    """
    Signal channels give the option control execution flow by triggering callback
    functions when the channel is called.
    Callbacks must be methods on the owner that require no positional arguments.
    Inputs optionally accept an output signal on call, which output signals always
    send when they call their input connections.

    Inputs hold a callback function to call, and outputs call each of their connections.

    Signal channels support `>>` as syntactic sugar for their connections, i.e.
    `some_output >> some_input` is equivalent to `some_input.connect(some_output)`.
    (This is also interoperable with `Node` objects.)
    """

    @abstractmethod
    def __call__(self) -> None:
        pass


class BadCallbackError(ValueError):
    pass


class InputSignal(SignalChannel["OutputSignal"], InputChannel["OutputSignal"]):

    def __init__(
        self,
        label: str,
        owner: Node,
        callback: typing.Callable,
    ):
        """
        Make a new input signal channel.

        Args:
            label (str): A name for the channel.
            owner (pyiron_workflow.node.Node): The channel's owner.
            callback (callable): An argument-free callback to invoke when calling this
                object. Must be a method on the owner.
        """
        super().__init__(label=label, owner=owner)
        if self._is_method_on_owner(callback) and self._all_args_arg_optional(callback):
            self._callback: str = callback.__name__
        else:
            raise BadCallbackError(
                f"The channel {self.full_label} got an unexpected callback: "
                f"{callback}. "
                f"Lives on owner: {self._is_method_on_owner(callback)}; "
                f"all args are optional: {self._all_args_arg_optional(callback)} "
            )

    @classmethod
    def connection_conjugate(cls) -> type[OutputSignal]:
        return OutputSignal

    def _is_method_on_owner(self, callback):
        try:
            return callback == getattr(self.owner, callback.__name__)
        except AttributeError:
            return False

    def _all_args_arg_optional(self, callback):
        return callable(callback) and not self._has_required_args(callback)

    @staticmethod
    def _has_required_args(func):
        return any(
            (
                param.kind
                in (
                    inspect.Parameter.POSITIONAL_ONLY,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    inspect.Parameter.KEYWORD_ONLY,
                )
                and param.default == inspect.Parameter.empty
            )
            for param in inspect.signature(func).parameters.values()
        )

    @property
    def callback(self) -> typing.Callable:
        return getattr(self.owner, self._callback)

    def __call__(self, other: OutputSignal | None = None) -> None:
        self.callback()

    def __str__(self):
        return f"{self.label} runs {self.callback.__name__}"

    def _connect_output_signal(self, signal: OutputSignal):
        self.connect(signal)


class AccumulatingInputSignal(InputSignal):
    """
    An input signal that only fires after receiving a signal from _all_ its connections
    instead of after _any_ of its connections.
    """

    def __init__(
        self,
        label: str,
        owner: Node,
        callback: typing.Callable,
    ):
        super().__init__(label=label, owner=owner, callback=callback)
        self.received_signals: set[str] = set()

    def __call__(self, other: OutputSignal | None = None) -> None:
        """
        Fire callback iff you have received at least one signal from each of your
        current connections.

        Resets the collection of received signals when firing.
        """
        if isinstance(other, OutputSignal):
            self.received_signals.update([other.scoped_label])
        if (
            len(
                {c.scoped_label for c in self.connections}.difference(
                    self.received_signals
                )
            )
            == 0
        ):
            self.reset()
            self.callback()

    def reset(self) -> None:
        """
        Reset the collection of received signals
        """
        self.received_signals = set()

    def __lshift__(self, others):
        others = others if isinstance(others, tuple) else (others,)
        for other in others:
            other._connect_accumulating_input_signal(self)


class OutputSignal(SignalChannel["InputSignal"], OutputChannel["InputSignal"]):

    @classmethod
    def connection_conjugate(cls) -> type[InputSignal]:
        return InputSignal

    def __call__(self) -> None:
        for c in self.connections:
            c(self)

    def __str__(self):
        return (
            f"{self.label} activates "
            f"{[f'{c.owner.label}.{c.label}' for c in self.connections]}"
        )

    def __rshift__(self, other: InputSignal | Node):
        other._connect_output_signal(self)
        return other

    def _connect_accumulating_input_signal(self, signal: AccumulatingInputSignal):
        self.connect(signal)



"""
A module to extract encapsulate for complex run mechanics, such as status, executor
interaction, etc.
"""

from __future__ import annotations

import contextlib
from abc import ABC, abstractmethod
from collections.abc import Callable
from concurrent.futures import Executor as StdLibExecutor
from concurrent.futures import Future, ThreadPoolExecutor
from functools import partial
from time import sleep
from typing import Any, TypeAlias

from pyiron_workflow.mixin.has_interface_mixins import HasLabel, HasRun, UsesState

InterpretableAsExecutor: TypeAlias = (
    StdLibExecutor | tuple[Callable[..., StdLibExecutor], tuple, dict]
)


class ReadinessError(ValueError):
    """
    To be raised when :class:`Runnable` calls run and requests a readiness check, but
    isn't :attr:`ready`.
    """

    readiness_dict: dict[str, bool]  # Detailed information on why it is not ready


class NotInterpretableAsExecutorError(TypeError): ...


class Runnable(UsesState, HasLabel, HasRun, ABC):
    """
    An abstract class for interfacing with executors, etc.

    Child classes must define :meth:`on_run` and :attr:`.Runnable.run_args`, then the
    :meth:`run` will invoke `self.on_run(*run_args[0], **run_args[1])`. The
    :class:`Runnable` class then handles the status of the run, passing the call off
    for remote execution, handling any returned futures object, etc.

    Child classes can optionally override :meth:`process_run_result` to do something
    with the returned value of :meth:`on_run`, but by default the returned value just
    passes cleanly through the function.

    The `run` cycle is broken down into sub-steps:
    - `_before_run`: prior to the `running` status being set to `True`
    - `_run`: after the `running` status has been set to `True`
    - `_finish_run`: what is done to the results of running, and when `running` is
        set to `False`
    - `_run_exception`: What to do if an encountered
    - `_run_finally`: What to do after _every_ run, regardless of whether an exception
        was encountered

    Child classes can extend the behavior of these sub-steps, including introducing
    new keyword arguments.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.running: bool = False
        self.failed: bool = False
        self._executor: InterpretableAsExecutor | None = None
        # We call it an executor, but it can also be instructions on making one
        self.future: None | Future = None
        self._thread_pool_sleep_time: float = 1e-6

    @abstractmethod
    def on_run(self, *args, **kwargs) -> Any:  # callable[..., Any | tuple]:
        """
        What the :meth:`run` method actually does!
        """

    @property
    @abstractmethod
    def run_args(self) -> tuple[tuple, dict]:
        """
        Any data needed for :meth:`on_run`, will be passed as (*args, **kwargs).
        """

    @property
    def executor(self) -> InterpretableAsExecutor | None:
        return self._executor

    @executor.setter
    def executor(self, executor: InterpretableAsExecutor | None):
        if not (
            isinstance(executor, StdLibExecutor | type(None))
            or (
                callable(executor[0])
                and isinstance(executor[1], tuple)
                and isinstance(executor[2], dict)
            )
        ):
            raise NotInterpretableAsExecutorError(
                f"Expected an instance of {StdLibExecutor}, or a tuple of such a "
                f"class, a tuple of args, and a dict of kwargs -- but got {executor}."
            )
        self._executor = executor

    def process_run_result(self, run_output: Any) -> Any:
        """
        What to _do_ with the results of :meth:`on_run` once you have them.

        By extracting this as a separate method, we allow the runnable to pass the
        actual execution off to another entity and release the python process to do
        other things. In such a case, this function should be registered as a callback
        so that the runnable can process the result of that process.

        Args:
            run_output: The results of a `self.on_run(self.run_args)` call.
        """
        return run_output

    @property
    def ready(self) -> bool:
        """Neither running nor failed"""
        return not (self.running or self.failed)

    @property
    def _readiness_dict(self) -> dict[str, bool]:
        return {
            "ready": self.ready,
            "running": self.running,
            "failed": self.failed,
        }

    @property
    def readiness_report(self) -> str:
        """A human-readable summary of the readiness to run."""
        report = f"{self.label} readiness report:\n"
        for k, v in self._readiness_dict.items():
            report += f"{k}: {v}\n"
        return report

    def executor_shutdown(self, wait=True, *, cancel_futures=False):
        """Invoke shutdown on the executor (if present)."""
        with contextlib.suppress(AttributeError):
            self.executor.shutdown(wait=wait, cancel_futures=cancel_futures)

    def run(
        self,
        check_readiness: bool = True,
        raise_run_exceptions: bool = True,
        rerun: bool = False,
        before_run_kwargs: dict | None = None,
        run_kwargs: dict | None = None,
        run_exception_kwargs: dict | None = None,
        run_finally_kwargs: dict | None = None,
        finish_run_kwargs: dict | None = None,
    ) -> Any | tuple | Future:
        """
        Checks that the runnable is :attr:`ready` (if requested), then executes the
        functionality of defined in :meth:`on_run` by passing it whatever is returned
        by :meth:`run_args`.

        Can stop early if :meth:`_before_run` called here returns `True` as its first
        argument.

        Handles the status of the runnable, communicating with any remote
        computing resources, and processing the result.

        Args:
            check_readiness (bool): Whether to raise a `ReadinessError` if not
                :attr:`ready`. (Default is True.)
            raise_run_exceptions (bool): Whether to raise exceptions encountered while
                :attr:`running`. (Default is True.)
            rerun (bool): Whether to proceed even if the :attr:`running` or
            :attr:`failed` state is encountered before runnign. (Default is False.)
        """

        def _none_to_dict(inp: dict | None) -> dict:
            return {} if inp is None else inp

        before_run_kwargs = _none_to_dict(before_run_kwargs)
        run_kwargs = _none_to_dict(run_kwargs)
        run_exception_kwargs = _none_to_dict(run_exception_kwargs)
        run_finally_kwargs = _none_to_dict(run_finally_kwargs)
        finish_run_kwargs = _none_to_dict(finish_run_kwargs)

        stop_early, result = self._before_run(
            check_readiness=check_readiness, rerun=rerun, **before_run_kwargs
        )
        if stop_early:
            return result

        self.running = True
        return self._run(
            raise_run_exceptions=raise_run_exceptions,
            run_exception_kwargs=run_exception_kwargs,
            run_finally_kwargs=run_finally_kwargs,
            finish_run_kwargs=finish_run_kwargs,
            **run_kwargs,
        )

    def _before_run(
        self, /, check_readiness: bool, rerun: bool, *args, **kwargs
    ) -> tuple[bool, Any]:
        """
        Things to do _before_ running.

        Args:
            check_readiness (bool): Whether to raise a `ReadinessError` if not
                :attr:`ready`.
            rerun (bool): Whether to proceed even if the :attr:`running` or
                :attr:`failed` state is encountered.
            **kwargs: Keyword arguments used by child classes in overriding this
                function.

        Returns:
            (bool): Whether to exit the parent run call early.
            (Any): What to return on an early-exit.

        Raises:
            (ReadinessError): If :param:`check_readiness` but not :attr:`ready`.
        """
        if rerun:
            self.running = False
            self.failed = False
        if check_readiness and not self.ready:
            readiness_error = ReadinessError(self._readiness_error_message)
            readiness_error.readiness_dict = self._readiness_dict
            raise readiness_error
        return False, None

    def _run(
        self,
        /,
        raise_run_exceptions: bool,
        run_exception_kwargs: dict,
        run_finally_kwargs: dict,
        finish_run_kwargs: dict,
        *args,
        **kwargs,
    ) -> Any | tuple | Future:
        """
        What happens while the status is :attr:`running`, namely invoking
        :meth:`self.on_run` using :attr:`self.run_args`, either locally or on an
        executor.

        Args:
            executor (concurrent.futures.Executor|None): Optionally, executor on which
                to run.
            raise_run_exceptions (bool): Whether to raise encountered exceptions.

        Returns:
            (Any | Future): The result of :meth:`on_run`, or a futures object from
                the executor.
        """
        on_run_args, on_run_kwargs = self.run_args
        if "self" in on_run_kwargs:
            raise ValueError(
                f"{self.label} got 'self' as a run kwarg, but self is already the "
                f"first positional argument passed to :meth:`on_run`."
            )

        if self.executor is None:
            try:
                run_output = self.on_run(*on_run_args, **on_run_kwargs)
            except (Exception, KeyboardInterrupt) as e:
                self._run_exception(**run_exception_kwargs)
                self._run_finally(**run_finally_kwargs)
                if raise_run_exceptions:
                    raise e
                else:
                    run_output = None
            return self._finish_run(
                run_output,
                raise_run_exceptions=raise_run_exceptions,
                run_exception_kwargs=run_exception_kwargs,
                run_finally_kwargs=run_finally_kwargs,
                unique_executor=None,
                **finish_run_kwargs,
            )
        else:
            if isinstance(self.executor, StdLibExecutor):
                executor = self.executor
                unique_executor = False
            else:
                creator, args, kwargs = self.executor
                executor = creator(*args, **kwargs)
                if not isinstance(executor, StdLibExecutor):
                    raise TypeError(
                        f"Expected an instance of {StdLibExecutor}, but got "
                        f"{type(executor)} from executor creation instructions "
                        f"{self.executor}."
                    )
                unique_executor = True

            submit_function = (
                self._thread_pool_run
                if isinstance(executor, ThreadPoolExecutor)
                else self.on_run
            )
            self.future = executor.submit(
                submit_function, *on_run_args, **on_run_kwargs
            )
            self.future.add_done_callback(
                partial(
                    self._finish_run,
                    raise_run_exceptions=raise_run_exceptions,
                    run_exception_kwargs=run_exception_kwargs,
                    run_finally_kwargs=run_finally_kwargs,
                    unique_executor=executor if unique_executor else None,
                    **finish_run_kwargs,
                )
            )

            return self.future

    def _run_exception(self, /, *args, **kwargs):
        """
        What to do if an exception is encountered inside :meth:`_run` or
        :meth:`_finish_run.
        """
        self.running = False
        self.failed = True

    def _run_finally(self, /, *args, **kwargs):
        """
        What to do after :meth:`_finish_run` (whether an exception is encountered or
        not), or in :meth:`_run` after an exception is encountered.
        """

    def _finish_run(
        self,
        run_output: tuple | Future,
        /,
        raise_run_exceptions: bool,
        run_exception_kwargs: dict,
        run_finally_kwargs: dict,
        unique_executor: StdLibExecutor | None,
        **kwargs,
    ) -> Any | tuple | None:
        """
        Switch the status, then process and return the run result.
        """
        self.running = False
        try:
            if isinstance(run_output, Future):
                run_output = run_output.result()
                self.future = None
                if unique_executor:
                    unique_executor.shutdown(wait=False)
                    del unique_executor
            return self.process_run_result(run_output)
        except Exception as e:
            self._run_exception(**run_exception_kwargs)
            if raise_run_exceptions:
                raise e
            return None
        finally:
            self._run_finally(**run_finally_kwargs)

    def _thread_pool_run(self, *args, **kwargs):
        """
        A poor attempt at avoiding (probably) thread races
        """
        result = self.on_run(*args, **kwargs)
        sleep(self._thread_pool_sleep_time)
        return result

    @property
    def _readiness_error_message(self) -> str:
        return (
            f"{self.label} received a run command but is not ready. The runnable "
            f"should be neither running nor failed.\n" + self.readiness_report
        )

    def __getstate__(self):
        state = super().__getstate__()
        state["future"] = None
        # Don't pass the future -- with the future in the state things work fine for
        # the simple pyiron_workflow.executors.CloudpickleProcessPoolExecutor, but for
        # the more complex executorlib.Executor we're getting:
        # TypeError: cannot pickle '_thread.RLock' object
        if isinstance(self._executor, StdLibExecutor):
            state["_executor"] = None
        # Don't pass actual executors, they have an unserializable thread lock on them
        # _but_ if the user is just passing instructions on how to _build_ an executor,
        # we'll trust that those serialize OK (this way we can, hopefully, eventually
        # support nesting executors!)
        return state



"""
We want to be able to operate freely on (single-output) nodes and channels as though
they were regular objects, but still retain our graph paradigm.

To accomplish this, we overload :class:`pyiron_workflow.channel.OutputData` to be able
to inject new nodes into the graph dynamically.
"""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any

from pyiron_workflow.channels import OutputData
from pyiron_workflow.data import NOT_DATA
from pyiron_workflow.io import GenericOutputs
from pyiron_workflow.mixin.has_interface_mixins import HasChannel

if TYPE_CHECKING:
    from pyiron_workflow.node import Node


class OutputDataWithInjection(OutputData):
    """
    Output data that must have a :class:`pyiron_workflow.node.Node` for its
    :attr:`owner`, and which is able to inject new nodes into that owner's graph, e.g.
    to accomplish operations on the channel.

    This class facilitates many (but not all) python operators by injecting a new
    node to perform that operation. Where the operator is not supported, we try to
    support using the operator's dunder name as a method, e.g. `==` gives us trouble
    with hashing, but this exploits the dunder method `.__eq__(other)`, so you can call
    `.eq(other)` on output data.
    These new nodes are instructed to run at the end of instantiation, but this fails
    cleanly in case they are not ready. This is intended to accommodate two likely
    scenarios: if you're injecting a node on top of an existing result you probably
    want the injection result to also be immediately available, but if you're injecting
    it at the end of something that hasn't run yet you don't want to see an error.
    """

    def __init__(
        self,
        label: str,
        owner: Node,
        default: Any | None = NOT_DATA,
        type_hint: Any | None = None,
        strict_hints: bool = True,
        value_receiver: OutputData | None = None,
    ):
        # Override parent method to give the new owner type hint
        super().__init__(
            label=label,
            owner=owner,
            default=default,
            type_hint=type_hint,
            strict_hints=strict_hints,
            value_receiver=value_receiver,
        )

    @staticmethod
    def _other_label(other):
        return (
            other.channel.scoped_label if isinstance(other, HasChannel) else str(other)
        )

    def _get_injection_label(self, injection_class, *args):
        other_labels = "_".join(self._other_label(other) for other in args)
        suffix = f"_{other_labels}" if len(args) > 0 else ""
        nominal_label = f"{self.scoped_label}_{injection_class.__name__}{suffix}"
        hashed = str(hash(nominal_label)).replace("-", "m")
        return f"injected_{injection_class.__name__}_{hashed}"

    def _node_injection(self, injection_class, *args, inject_self=True):
        """
        Create a new node with the same parent as this channel's node, and feed it
        arguments, or load such a node if it already exists on the parent (based on a
        name dynamically generated from the injected node class and arguments).

        Args:
            injection_class (type[Node]): The new node class to instantiate
            *args: Any arguments for that function node
            inject_self (bool): Whether to pre-pend the args with self. (Default is
                True.)

        Returns:
            (Node): The instantiated or loaded node.
        """
        label = self._get_injection_label(injection_class, *args)
        try:
            # First check if the node already exists
            return self.owner.parent.children[label]
        except (AttributeError, KeyError):
            # Fall back on creating a new node in case parent is None or node nexists
            node_args = (self, *args) if inject_self else args
            return injection_class(
                *node_args, parent=self.owner.parent, label=label, autorun=True
            )

    # We don't wrap __all__ the operators, because you might really want the string or
    # hash or whatever of the actual channel. But we do wrap all the dunder methods
    # that should be unambiguously referring to an operation on values

    def __getattr__(self, name):
        from pyiron_workflow.nodes.standard import GetAttr

        if name == "to_hdf":
            raise AttributeError(
                "This is just a failsafe to protect us against other elements of the "
                "pyiron ecosystem (pyiron_base's DataContainer) running a "
                "`hasattr('to_hdf')` check on us and accidentally injecting a new "
                "getattr node."
            )
        if name.startswith("_"):
            raise AttributeError(
                f"{self.full_label} ({OutputDataWithInjection.__name__}) tried to "
                f"inject on the attribute {name}, but injecting on private attributes "
                f"is forbidden -- if you really need it create a {GetAttr.__name__} "
                f"node manually."
            )
        if name == "shape":
            raise AttributeError(
                "This is a hack to stop jupyter notebook cells from asking for a `shape`."
                "If you are _actually_ trying to get delayed access to a `shape` field on "
                "your output, you'll need to manually add an attribute access node to do "
                "it."
            )
        return self._node_injection(GetAttr, name)

    def __getitem__(self, item):
        # Break slices into deeper injections, if any slice arguments are channel-like
        if isinstance(item, slice) and any(
            isinstance(slice_input, HasChannel)
            for slice_input in [item.start, item.stop, item.step]
        ):
            from pyiron_workflow.nodes.standard import Slice

            item = self._node_injection(
                Slice, item.start, item.stop, item.step, inject_self=False
            )

        from pyiron_workflow.nodes.standard import GetItem

        return self._node_injection(GetItem, item)

    def __lt__(self, other):
        from pyiron_workflow.nodes.standard import LessThan

        return self._node_injection(LessThan, other)

    def __le__(self, other):
        from pyiron_workflow.nodes.standard import LessThanEquals

        return self._node_injection(LessThanEquals, other)

    def eq(self, other):
        from pyiron_workflow.nodes.standard import Equals

        return self._node_injection(Equals, other)

    def __ne__(self, other):
        from pyiron_workflow.nodes.standard import NotEquals

        return self._node_injection(NotEquals, other)

    def __gt__(self, other):
        from pyiron_workflow.nodes.standard import GreaterThan

        return self._node_injection(GreaterThan, other)

    def __ge__(self, other):
        from pyiron_workflow.nodes.standard import GreaterThanEquals

        return self._node_injection(GreaterThanEquals, other)

    def bool(self):
        from pyiron_workflow.nodes.standard import Bool

        return self._node_injection(Bool)

    def len(self):
        from pyiron_workflow.nodes.standard import Length

        return self._node_injection(Length)

    def contains(self, other):
        from pyiron_workflow.nodes.standard import Contains

        return self._node_injection(Contains, other)

    def __add__(self, other):
        from pyiron_workflow.nodes.standard import Add

        return self._node_injection(Add, other)

    def __sub__(self, other):
        from pyiron_workflow.nodes.standard import Subtract

        return self._node_injection(Subtract, other)

    def __mul__(self, other):
        from pyiron_workflow.nodes.standard import Multiply

        return self._node_injection(Multiply, other)

    def __rmul__(self, other):
        from pyiron_workflow.nodes.standard import RightMultiply

        return self._node_injection(RightMultiply, other)

    def __matmul__(self, other):
        from pyiron_workflow.nodes.standard import MatrixMultiply

        return self._node_injection(MatrixMultiply, other)

    def __truediv__(self, other):
        from pyiron_workflow.nodes.standard import Divide

        return self._node_injection(Divide, other)

    def __floordiv__(self, other):
        from pyiron_workflow.nodes.standard import FloorDivide

        return self._node_injection(FloorDivide, other)

    def __mod__(self, other):
        from pyiron_workflow.nodes.standard import Modulo

        return self._node_injection(Modulo, other)

    def __pow__(self, other):
        from pyiron_workflow.nodes.standard import Power

        return self._node_injection(Power, other)

    def __and__(self, other):
        from pyiron_workflow.nodes.standard import And

        return self._node_injection(And, other)

    def __xor__(self, other):
        from pyiron_workflow.nodes.standard import XOr

        return self._node_injection(XOr, other)

    def __or__(self, other):
        from pyiron_workflow.nodes.standard import Or

        return self._node_injection(Or, other)

    def __neg__(self):
        from pyiron_workflow.nodes.standard import Negative

        return self._node_injection(Negative)

    def __pos__(self):
        from pyiron_workflow.nodes.standard import Positive

        return self._node_injection(Positive)

    def __abs__(self):
        from pyiron_workflow.nodes.standard import Absolute

        return self._node_injection(Absolute)

    def __invert__(self):
        from pyiron_workflow.nodes.standard import Invert

        return self._node_injection(Invert)

    def int(self):
        from pyiron_workflow.nodes.standard import Int

        return self._node_injection(Int)

    def float(self):
        from pyiron_workflow.nodes.standard import Float

        return self._node_injection(Float)

    def __round__(self):
        from pyiron_workflow.nodes.standard import Round

        return self._node_injection(Round)


class OutputsWithInjection(GenericOutputs[OutputDataWithInjection]):
    @property
    def _channel_class(self) -> type[OutputDataWithInjection]:
        return OutputDataWithInjection


class InjectsOnChannel(HasChannel, abc.ABC):
    @property
    @abc.abstractmethod
    def channel(self) -> OutputDataWithInjection: ...

    def __getattr__(self, item):
        try:
            return super().__getattr__(item)
        except AttributeError:
            channel = self.channel
            return getattr(channel, item)

    def __getitem__(self, item):
        return self.channel.__getitem__(item)

    def __lt__(self, other):
        return self.channel.__lt__(other)

    def __le__(self, other):
        return self.channel.__le__(other)

    def eq(self, other):
        return self.channel.eq(other)

    def __ne__(self, other):
        return self.channel.__ne__(other)

    def __gt__(self, other):
        return self.channel.__gt__(other)

    def __ge__(self, other):
        return self.channel.__ge__(other)

    def bool(self):
        return self.channel.bool()

    def len(self):
        return self.channel.len()

    def contains(self, other):
        return self.channel.contains(other)

    def __add__(self, other):
        return self.channel.__add__(other)

    def __sub__(self, other):
        return self.channel.__sub__(other)

    def __mul__(self, other):
        return self.channel.__mul__(other)

    def __rmul__(self, other):
        return self.channel.__rmul__(other)

    def __matmul__(self, other):
        return self.channel.__matmul__(other)

    def __truediv__(self, other):
        return self.channel.__truediv__(other)

    def __floordiv__(self, other):
        return self.channel.__floordiv__(other)

    def __mod__(self, other):
        return self.channel.__mod__(other)

    def __pow__(self, other):
        return self.channel.__pow__(other)

    def __and__(self, other):
        return self.channel.__and__(other)

    def __xor__(self, other):
        return self.channel.__xor__(other)

    def __or__(self, other):
        return self.channel.__or__(other)

    def __neg__(self):
        return self.channel.__neg__()

    def __pos__(self):
        return self.channel.__pos__()

    def __abs__(self):
        return self.channel.__abs__()

    def __invert__(self):
        return self.channel.__invert__()

    def int(self):
        return self.channel.int()

    def float(self):
        return self.channel.float()

    def __round__(self):
        return self.channel.__round__()



"""
Complete API for users and node developers.

All public items included here are subject to semantic versioning.
"""

# Node developer entry points
from pyiron_workflow.data import NOT_DATA
from pyiron_workflow.executors import CloudpickleProcessPoolExecutor, NodeSlurmExecutor
from pyiron_workflow.find import (
    find_nodes as _find_nodes,  # Not formally in API -- don't rely on interface
)
from pyiron_workflow.logging import logger
from pyiron_workflow.nodes import standard as std  # User resource
from pyiron_workflow.nodes.composite import FailedChildError
from pyiron_workflow.nodes.for_loop import For, for_node, for_node_factory
from pyiron_workflow.nodes.function import (
    Function,
    as_function_node,
    function_node,
    to_function_node,
)
from pyiron_workflow.nodes.macro import Macro, as_macro_node, macro_node
from pyiron_workflow.nodes.transform import (
    as_dataclass_node,
    dataclass_node,
    inputs_to_dataframe,
    inputs_to_dict,
    inputs_to_list,
    list_to_outputs,
)
from pyiron_workflow.nodes.while_loop import While, while_node, while_node_factory
from pyiron_workflow.storage import (
    PickleStorage,
    StorageInterface,
    TypeNotFoundError,
    available_backends,
)
from pyiron_workflow.workflow import Workflow



"""
Functions for drawing the graph.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, Literal, TypeVar

import graphviz
from pyiron_snippets.colors import SeabornColors

from pyiron_workflow.data import NOT_DATA

if TYPE_CHECKING:
    from pyiron_workflow.channels import Channel as WorkflowChannel  # noqa: F401
    from pyiron_workflow.channels import (
        DataChannel as WorkflowDataChannel,  # noqa: F401
    )
    from pyiron_workflow.channels import (
        SignalChannel as WorkflowSignalChannel,  # noqa: F401
    )
    from pyiron_workflow.io import DataIO, SignalIO
    from pyiron_workflow.node import Node as WorkflowNode


def directed_graph(
    name,
    label,
    rankdir,
    color_start,
    color_end,
    gradient_angle,
    size=None,
    style="filled",
    fillcolor_start=None,
    fillcolor_end=None,
):
    """A shortcut method for instantiating the type of graphviz graph we want"""
    if fillcolor_start is None:
        fillcolor_start = color_start
    if fillcolor_end is None:
        fillcolor_end = color_end
    digraph = graphviz.graphs.Digraph(name=name)
    digraph.attr(
        label=label,
        compound="true",
        rankdir=rankdir,
        style=style,
        fillcolor=f"{fillcolor_start}:{fillcolor_end}",
        color=f"{color_start}:{color_end}",
        gradientangle=gradient_angle,
        fontname="helvetica",
        size=size,
    )
    return digraph


def reverse_rankdir(rankdir: Literal["LR", "TB"]):
    if rankdir == "LR":
        return "TB"
    elif rankdir == "TB":
        return "LR"
    else:
        raise ValueError(f"Expected rankdir of 'LR' or 'TB' but got {rankdir}")


def _to_hex(rgb: tuple[int, int, int]) -> str:
    """RGB [0,255] to hex color codes; no alpha values."""
    return "#{:02x}{:02x}{:02x}".format(*tuple(int(c) for c in rgb))


def _to_rgb(hex_: str) -> tuple[int, int, int]:
    """Hex to RGB color codes; no alpha values."""
    hex_ = hex_.lstrip("#")
    return (
        int(hex_[0 : 0 + 2], 16),
        int(hex_[2 : 2 + 2], 16),
        int(hex_[4 : 4 + 2], 16),
    )  # mypy isn't smart enough to parse this as a 3-tuple from an iterator


def blend_colours(color_a, color_b, fraction_a=0.5):
    """Blends two hex code colours together"""
    return _to_hex(
        tuple(
            fraction_a * a + (1 - fraction_a) * b
            for (a, b) in zip(_to_rgb(color_a), _to_rgb(color_b), strict=False)
        )
    )


def lighten_hex_color(color, lightness=0.7):
    """Blends the given hex code color with pure white."""
    return blend_colours(SeabornColors.white, color, fraction_a=lightness)


class WorkflowGraphvizMap(ABC):
    """
    A parent class defining the interface for the graphviz representation of all our
    workflow objects.
    """

    @property
    @abstractmethod
    def parent(self) -> WorkflowGraphvizMap | None:
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def label(self) -> str:
        pass

    @property
    @abstractmethod
    def graph(self) -> graphviz.graphs.Digraph:
        pass

    @property
    @abstractmethod
    def color(self) -> str:
        pass


WorkflowChannelType = TypeVar("WorkflowChannelType", bound="WorkflowChannel")


class _Channel(WorkflowGraphvizMap, Generic[WorkflowChannelType], ABC):
    """
    An abstract representation for channel objects, which are "nodes" in graphviz
    parlance.
    """

    def __init__(self, parent: _IO, channel: WorkflowChannelType, local_name: str):
        self.channel: WorkflowChannelType = channel
        self._parent = parent
        self._name = self.parent.name + local_name
        self._label = local_name + self._build_label_suffix()

        self.graph.node(
            name=self.name,
            label=self.label,
            shape=self.shape,
            color=self.color,
            style=self.style,
            fontname="helvetica",
        )

    @property
    @abstractmethod
    def shape(self) -> str:
        pass

    def _build_label_suffix(self):
        suffix = ""
        try:
            if self.channel.type_hint is not None:
                suffix += ": " + self.channel.type_hint.__name__
        except AttributeError:
            pass  # Signals have no type
        return suffix

    @property
    def parent(self) -> _IO:
        return self._parent

    @property
    def name(self) -> str:
        return self._name

    @property
    def label(self) -> str:
        return self._label

    @property
    def graph(self) -> graphviz.graphs.Digraph:
        return self.parent.graph

    @property
    def style(self) -> str:
        return "filled"


class DataChannel(_Channel["WorkflowDataChannel"]):
    @property
    def color(self) -> str:
        orange = "#EDB22C"
        return orange

    @property
    def shape(self) -> str:
        return "oval"

    @property
    def style(self) -> str:
        if len(self.channel.connections) == 0 and self.channel.value is NOT_DATA:
            return "bold"
        return "filled"


class SignalChannel(_Channel["WorkflowSignalChannel"]):
    @property
    def color(self) -> str:
        blue = "#21BFD8"
        return blue

    @property
    def shape(self) -> str:
        return "cds"


class _IO(WorkflowGraphvizMap, ABC):
    """
    An abstract class for IO panels, which are represented as a "subgraph" in graphviz
    parlance.
    """

    def __init__(self, parent: Node):
        self._parent = parent
        self.node: WorkflowNode = self.parent.node
        self.data_io, self.signals_io = self._get_node_io()
        self._name = self.parent.name + self.data_io.__class__.__name__
        self._label = self.data_io.__class__.__name__
        self._graph = directed_graph(
            self.name,
            self.label,
            rankdir=reverse_rankdir(self.parent.rankdir),
            color_start=self.color,
            color_end=lighten_hex_color(self.color),
            gradient_angle=self.gradient_angle,
        )

        self.channels = [
            SignalChannel(self, channel, panel_label)
            for panel_label, channel in self.signals_io.items()
        ] + [
            DataChannel(self, channel, panel_label)
            for panel_label, channel in self.data_io.items()
        ]

        self.parent.graph.subgraph(self.graph)

    @abstractmethod
    def _get_node_io(self) -> tuple[DataIO, SignalIO]:
        pass

    @property
    @abstractmethod
    def gradient_angle(self) -> str:
        """Background fill colour angle in degrees"""

    @property
    def parent(self) -> Node:
        return self._parent

    @property
    def name(self) -> str:
        return self._name

    @property
    def label(self) -> str:
        return self._label

    @property
    def graph(self) -> graphviz.graphs.Digraph:
        return self._graph

    @property
    def color(self) -> str:
        gray = "#A5A4A5"
        return gray

    def __len__(self):
        return len(self.channels)


class Inputs(_IO):
    def _get_node_io(self) -> tuple[DataIO, SignalIO]:
        return self.node.inputs, self.node.signals.input

    @property
    def gradient_angle(self) -> str:
        return "0"


class Outputs(_IO):
    def _get_node_io(self) -> tuple[DataIO, SignalIO]:
        return self.node.outputs, self.node.signals.output

    @property
    def gradient_angle(self) -> str:
        return "180"


class Node(WorkflowGraphvizMap):
    """
    A wrapper class to connect graphviz to our workflow nodes. The nodes are
    represented by a "graph" or "subgraph" in graphviz parlance (depending on whether
    the node being visualized is the top-most node or not).

    Visualized nodes show their label and type, and IO panels with label and type.
    Colors and shapes are exploited to differentiate various node classes, input/output,
    and data/signal channels.

    If the node is composite in nature and the `depth` argument is at least `1`, owned
    children are also visualized (recursively with `depth = depth - 1`) inside the scope
    of this node.

    Args:
        node (pyiron_workflow.node.Node): The node to visualize.
        parent (Optional[pyiron_workflow.draw.Node]): The visualization that
            owns this visualization (if any).
        depth (int): How deeply to decompose any child nodes beyond showing their IO.
        rankdir ("LR" | "TB"): Use left-right or top-bottom graphviz `rankdir`.
        size (tuple[int | float, int | float] | None): The size of the diagram, in
            inches(?); respects ratio by scaling until at least one dimension matches
            the requested size. (Default is None, automatically size.)
    """

    def __init__(
        self,
        node: WorkflowNode,
        parent: Node | None = None,
        depth: int = 1,
        rankdir: Literal["LR", "TB"] = "LR",
        size: str | None = None,
    ):
        self.node = node
        self._parent = parent
        self._name = self.build_node_name()
        self._label = self.node.label + ": " + self.node.__class__.__name__
        self.rankdir: Literal["LR", "TB"] = rankdir
        self._graph = directed_graph(
            self.name,
            self.label,
            rankdir=self.rankdir,
            color_start=self.color,
            color_end=self.color,
            gradient_angle="0",
            size=size,
            style=self.style,
            fillcolor_start=self.fillcolor,
            fillcolor_end=self.fillcolor,
        )

        self.inputs = Inputs(self)
        self.outputs = Outputs(self)
        self.graph.edge(
            self.inputs.channels[0].name, self.outputs.channels[0].name, style="invis"
        )

        if depth > 0:
            from pyiron_workflow.nodes.composite import Composite  # noqa: PLC0415

            # Janky in-line import to avoid circular imports but only look for children
            # where they exist (since nodes sometimes now actually do something on
            # failed attribute access, i.e. use it as delayed access on their output)
            if isinstance(self.node, Composite):
                self._connect_owned_nodes(depth)

        if self.parent is not None:
            self.parent.graph.subgraph(self.graph)

    def _channel_bicolor(self, start_channel, end_channel):
        return f"{start_channel.color};0.5:{end_channel.color};0.5"

    def _connect_owned_nodes(self, depth):
        nodes = [Node(node, self, depth - 1) for node in self.node.children.values()]
        internal_inputs = [
            channel for node in nodes for channel in node.inputs.channels
        ]
        internal_outputs = [
            channel for node in nodes for channel in node.outputs.channels
        ]

        # Loop to check for internal node output --> internal node input connections
        for output_channel in internal_outputs:
            for input_channel in internal_inputs:
                if input_channel.channel in output_channel.channel.connections:
                    self.graph.edge(
                        output_channel.name,
                        input_channel.name,
                        color=self._channel_bicolor(output_channel, input_channel),
                    )

        # Connect channels that are by-reference copies of each other
        # i.e. for Workflow IO to child IO
        self._connect_matching(self.inputs.channels, internal_inputs)
        self._connect_matching(internal_outputs, self.outputs.channels)

        # Connect channels that are value-linked
        # i.e. for Macro IO to child IO
        self._connect_linked(self.inputs.channels, internal_inputs)
        self._connect_linked(internal_outputs, self.outputs.channels)

    def _connect_matching(self, sources: list[_Channel], destinations: list[_Channel]):
        """
        Draw an edge between two graph channels whose workflow channels are the same
        """
        for source in sources:
            for destination in destinations:
                if source.channel is destination.channel:
                    self.graph.edge(
                        source.name,
                        destination.name,
                        color=self._channel_bicolor(source, destination),
                    )

    def _connect_linked(self, sources: list[_Channel], destinations: list[_Channel]):
        """
        Draw an edge between two graph channels values are linked
        """
        for source in sources:
            for destination in destinations:
                if (
                    hasattr(source.channel, "value_receiver")
                    and source.channel.value_receiver is destination.channel
                ):
                    self.graph.edge(
                        source.name,
                        destination.name,
                        color=self._channel_bicolor(source, destination),
                        style="dashed",
                    )

    def build_node_name(self, suffix=""):
        if self.parent is not None:
            # Recursively prepend parent labels to get a totally unique label string
            # (inside the scope of this graph)
            return self.parent.build_node_name(suffix=suffix + self.node.label)
        else:
            return "cluster" + self.node.label + suffix

    @property
    def parent(self) -> Node | None:
        return self._parent

    @property
    def name(self) -> str:
        return self._name

    @property
    def label(self) -> str:
        return self._label

    @property
    def graph(self) -> graphviz.graphs.Digraph:
        return self._graph

    @property
    def color(self) -> str:
        bright_red = "#FF0000"
        if self.node.failed:
            return bright_red
        elif self.node.running:
            return SeabornColors.orange
        elif self.node.cache_hit and self.node.outputs.ready:  # Successfully ran
            return SeabornColors.green
        else:
            return self.fillcolor

    @property
    def fillcolor(self) -> str:
        return lighten_hex_color(self.node.color)

    @property
    def style(self) -> str:
        return "filled, bold"



"""
Common-use nodes relying only on the standard library
"""

from __future__ import annotations

import os
import random
import shutil
from collections.abc import Callable
from pathlib import Path
from time import sleep

from pyiron_workflow.channels import OutputSignal
from pyiron_workflow.data import NOT_DATA
from pyiron_workflow.nodes.function import Function, as_function_node


@as_function_node
def UserInput(user_input):
    """
    Returns the user input as it is.

    Args:
        user_input: Any input provided by the user.

    Returns:
        The same input provided by the user.
    """
    return user_input


class If(Function):
    """
    Has two extra signal channels: true and false. Evaluates the input as a boolean and
    fires the corresponding output signal after running.

    Args:
        **kwargs: Additional keyword arguments for the Function base class.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.signals.output.true = OutputSignal("true", self)
        self.signals.output.false = OutputSignal("false", self)

    @staticmethod
    def node_function(condition):
        """
        Evaluate the condition as a boolean.

        Args:
            condition: The condition to evaluate.

        Returns:
            bool: The boolean value of the condition.

        Raises:
            TypeError: If the condition is NOT_DATA.
        """
        if condition is NOT_DATA:
            raise TypeError("Logic 'If' node expected data but got NOT_DATA as input.")
        truth = bool(condition)
        return truth

    @property
    def emitting_channels(self) -> tuple[OutputSignal, ...]:
        if self.outputs.truth.value is NOT_DATA:
            return super().emitting_channels
        elif self.outputs.truth.value:
            return (*super().emitting_channels, self.signals.output.true)
        else:
            return (*super().emitting_channels, self.signals.output.false)


@as_function_node("list")
def AppendToList(existing: list | None = None, new_element=NOT_DATA):
    """
    Append a new element to a list.

    Args:
        existing (list | None): The list to append to. Defaults to None, which creates
            an empty list.
        new_element: The new element to append. This is mandatory input as the default
            value of `NOT_DATA` will cause a readiness error.

    Returns:
        list: The input list with the new element appended

    Examples:
        This node is particularly useful for acyclic flows, where you want to bend the
        suggestions of idempotency by looping the node's output back on itself:

        >>> from pyiron_workflow import std
        >>>
        >>> n = std.AppendToList()
        >>> n.run(new_element="foo")
        ['foo']

        >>> n.run(existing=n, new_element="bar")
        ['foo', 'bar']

        >>> n.run(existing=n, new_element="baz")
        ['foo', 'bar', 'baz']

    """
    existing = [] if existing is None else existing
    if new_element is not NOT_DATA:
        existing.append(new_element)
    return existing


@as_function_node("started_at", "ended_at", "was_empty")
def ChangeDirectory(
    path: str | Path,
    delete_start: bool = False,
) -> tuple[str, str, bool]:
    """
    Change directory.

    Args:
        path (str | Path): Where to go.
        delete_start (bool): Whether to recursively remove the starting location. (Yes,
            this is dangerous, be careful -- but it's useful for going in and out of
            temp directories.)

    Returns:
        str: The absolute path to where we started.

    Examples:
        All the extra input and output allows use to go in and out of a location
        leaving no trace if we created it to start with.
        >>> import os
        >>>
        >>> from pyiron_workflow import std
        >>>
        >>> go_in = std.ChangeDirectory(path="some_subdirectory")
        >>> in_result = go_in.run()
        >>> print(
        ...     os.getcwd().replace(
        ...         go_in.outputs.started_at.value, "some path..."
        ...     ).replace(
        ...          os.path.sep, "/"  # For doctesting on windows...
        ...     )
        ... )
        some path.../some_subdirectory

        >>> go_out = std.ChangeDirectory(
        ...     path=go_in.outputs.started_at,
        ...     delete_start=go_in.outputs.was_empty
        ... )
        >>> out_result = go_out.run();
        >>> print(os.getcwd() == go_in.outputs.started_at.value)
        True

        >>> print(os.path.isdir("some_subdirectory"))
        False

    """
    started_at = os.getcwd()
    if isinstance(path, Path):
        path = str(path.resolve())
    os.makedirs(path, exist_ok=True)
    was_empty = not any(os.scandir(path))
    os.chdir(path)
    if delete_start:
        shutil.rmtree(started_at)
    return started_at, path, was_empty


@as_function_node
def PureCall(fnc: Callable):
    """
    Return a call without any arguments

    Args:
        fnc (callable): The callable object.

    Returns:
        An argument-free call on the callable object.

    Examples:
        This is particularly useful for snagging methods off objects inside a workflow.
        >>> import datetime
        >>>
        >>> from pyiron_workflow import std
        >>>
        >>> inp = std.UserInput(datetime.date(1977, 5, 25))
        >>> pure_call = std.PureCall(inp.isoformat)
        >>> pure_call()
        '1977-05-25'

        Where we got the method to call using an injected `GetAttr` node.
    """
    return fnc()


@as_function_node("random")
def RandomFloat():
    """
    Generates a random float between 0 and 1.

    Returns:
        float: A random float between 0 and 1.
    """
    return random.random()


@as_function_node("time")
def Sleep(t):
    """
    Sleeps for the given number of seconds.

    Args:
        t (float): Number of seconds to sleep.

    Returns:
        float: The same number of seconds slept.
    """
    sleep(t)
    return t


@as_function_node("slice")
def Slice(start=None, stop=NOT_DATA, step=None):
    """
    Creates a slice object.

    Args:
        start (int | None): The start index. If None (Default), slicing starts from the
            beginning.
        stop (int | None): The stop index. If None or NOT_DATA (Default), slicing goes
            until the end.
        step (int | None): The step index. If None (Default), slicing proceeds with a
            step of 1.

    Returns:
        slice: The created slice object.

    Raises:
        ValueError: If the arguments are not valid for creating a slice.
    """
    if start is None:
        if stop is None:
            raise ValueError(
                "Slice must define at least start or stop, but both are None"
            )
        elif step is not None:
            raise ValueError("If step is provided, start _must_ be provided")
        else:
            s = slice(stop)
    elif stop is None:
        raise ValueError("If start is provided, stop _must_ be provided")
    else:
        s = slice(start, stop, step)
    return s


@as_function_node("object")
def SetAttr(obj, key: str, val):
    """
    Sets an attribute on an object.

    Args:
        obj: The object on which to set the attribute.
        key (str): The attribute name.
        val: The value to set for the attribute.

    Returns:
        The object with the attribute set.
    """
    setattr(obj, key, val)
    return obj


@as_function_node("str")
def String(obj):
    """
    Converts an object to its string representation.

    Args:
        obj: The object to convert.

    Returns:
        str: The string representation of the object.
    """
    return str(obj)


@as_function_node("bytes")
def Bytes(obj):
    """
    Converts an object to its bytes representation.

    Args:
        obj: The object to convert.

    Returns:
        bytes: The bytes representation of the object.
    """
    return bytes(obj)


@as_function_node("lt")
def LessThan(obj, other):
    """
    Compares if obj is less than other.

    Args:
        obj: The first object to compare.
        other: The second object to compare.

    Returns:
        bool: True if obj is less than other, False otherwise.
    """
    return obj < other


@as_function_node("le")
def LessThanEquals(obj, other):
    """
    Compares if obj is less than or equal to other.

    Args:
        obj: The first object to compare.
        other: The second object to compare.

    Returns:
        bool: True if obj is less than or equal to other, False otherwise.
    """
    return obj <= other


@as_function_node("eq")
def Equals(obj, other):
    """
    Compares if obj is equal to other.

    Args:
        obj: The first object to compare.
        other: The second object to compare.

    Returns:
        bool: True if obj is equal to other, False otherwise.
    """
    return obj == other


@as_function_node("neq")
def NotEquals(obj, other):
    """
    Compares if obj is not equal to other.

    Args:
        obj: The first object to compare.
        other: The second object to compare.

    Returns:
        bool: True if obj is not equal to other, False otherwise.
    """
    return obj != other


@as_function_node("gt")
def GreaterThan(obj, other):
    """
    Compares if obj is greater than other.

    Args:
        obj: The first object to compare.
        other: The second object to compare.

    Returns:
        bool: True if obj is greater than other, False otherwise.
    """
    return obj > other


@as_function_node("ge")
def GreaterThanEquals(obj, other):
    """
    Compares if obj is greater than or equal to other.

    Args:
        obj: The first object to compare.
        other: The second object to compare.

    Returns:
        bool: True if obj is greater than or equal to other, False otherwise.
    """
    return obj >= other


@as_function_node("hash")
def Hash(obj):
    """
    Returns the hash value of an object.

    Args:
        obj: The object to hash.

    Returns:
        int: The hash value of the object.
    """
    return hash(obj)


@as_function_node("bool")
def Bool(obj):
    """
    Converts an object to its boolean representation.

    Args:
        obj: The object to convert.

    Returns:
        bool: The boolean representation of the object.
    """
    return bool(obj)


@as_function_node("getattr")
def GetAttr(obj, name):
    """
    Gets an attribute from an object.

    Args:
        obj: The object from which to get the attribute.
        name: The name of the attribute.

    Returns:
        The value of the attribute.
    """
    return getattr(obj, name)


@as_function_node("getitem")
def GetItem(obj, item):
    """
    Gets an item from an object.

    Args:
        obj: The object from which to get the item.
        item: The item to get.

    Returns:
        The value of the item.
    """
    return obj[item]


@as_function_node("dir")
def Dir(obj):
    """
    Returns a list of valid attributes for the object.

    Args:
        obj: The object to inspect.

    Returns:
        list: A list of valid attributes for the object.
    """
    return dir(obj)


@as_function_node("len")
def Length(obj):
    """
    Returns the length of an object.

    Args:
        obj: The object to measure.

    Returns:
        int: The length of the object.
    """
    return len(obj)


@as_function_node("in")
def Contains(obj, other):
    """
    Checks if obj contains other.

    Args:
        obj: The object to check.
        other: The item to check for.

    Returns:
        bool: True if obj contains other, False otherwise.
    """
    return other in obj


@as_function_node("add")
def Add(obj, other):
    """
    Adds obj and other.

    Args:
        obj: The first operand.
        other: The second operand.

    Returns:
        The result of adding obj and other.
    """
    return obj + other


@as_function_node("sub")
def Subtract(obj, other):
    """
    Subtracts other from obj.

    Args:
        obj: The first operand.
        other: The second operand.

    Returns:
        The result of subtracting other from obj.
    """
    return obj - other


@as_function_node("mul")
def Multiply(obj, other):
    """
    Multiplies obj by other.

    Args:
        obj: The first operand.
        other: The second operand.

    Returns:
        The result of multiplying obj by other.
    """
    return obj * other


@as_function_node("rmul")
def RightMultiply(obj, other):
    """
    Multiplies other by obj (reversed operands).

    Args:
        obj: The second operand.
        other: The first operand.

    Returns:
        The result of multiplying other by obj.
    """
    return other * obj


@as_function_node("matmul")
def MatrixMultiply(obj, other):
    """
    Performs matrix multiplication on obj and other.

    Args:
        obj: The first matrix.
        other: The second matrix.

    Returns:
        The result of the matrix multiplication.
    """
    return obj @ other


@as_function_node("truediv")
def Divide(obj, other):
    """
    Divides obj by other.

    Args:
        obj: The numerator.
        other: The denominator.

    Returns:
        The result of dividing obj by other.
    """
    return obj / other


@as_function_node("floordiv")
def FloorDivide(obj, other):
    """
    Performs floor division on obj by other.

    Args:
        obj: The numerator.
        other: The denominator.

    Returns:
        The result of floor division.
    """
    return obj // other


@as_function_node("mod")
def Modulo(obj, other):
    """
    Calculates the modulo of obj by other.

    Args:
        obj: The numerator.
        other: The denominator.

    Returns:
        The result of obj modulo other.
    """
    return obj % other


@as_function_node("pow")
def Power(obj, other):
    """
    Raises obj to the power of other.

    Args:
        obj: The base.
        other: The exponent.

    Returns:
        The result of obj raised to the power of other.
    """
    return obj**other


@as_function_node("and")
def And(obj, other):
    """
    Performs a bitwise AND operation on obj and other.

    Args:
        obj: The first operand.
        other: The second operand.

    Returns:
        The result of the bitwise AND operation.
    """
    return obj & other


@as_function_node("xor")
def XOr(obj, other):
    """
    Performs a bitwise XOR operation on obj and other.

    Args:
        obj: The first operand.
        other: The second operand.

    Returns:
        The result of the bitwise XOR operation.
    """
    return obj ^ other


@as_function_node("or")
def Or(obj, other):
    """
    Performs a bitwise OR operation on obj and other.

    Args:
        obj: The first operand.
        other: The second operand.

    Returns:
        The result of the bitwise OR operation.
    """
    return obj | other


@as_function_node("neg")
def Negative(obj):
    """
    Negates an object.

    Args:
        obj: The object to negate.

    Returns:
        The negated object.
    """
    return -obj


@as_function_node("pos")
def Positive(obj):
    """
    Returns the positive of an object.

    Args:
        obj: The object.

    Returns:
        The positive of the object.
    """
    return +obj


@as_function_node("abs")
def Absolute(obj):
    """
    Returns the absolute value of an object.

    Args:
        obj: The object.

    Returns:
        The absolute value of the object.
    """
    return abs(obj)


@as_function_node("invert")
def Invert(obj):
    """
    Inverts the bits of an object.

    Args:
        obj: The object.

    Returns:
        The inverted object.
    """
    return ~obj


@as_function_node("int")
def Int(obj):
    """
    Converts an object to an integer.

    Args:
        obj: The object to convert.

    Returns:
        int: The integer representation of the object.
    """
    return int(obj)


@as_function_node("float")
def Float(obj):
    """
    Converts an object to a float.

    Args:
        obj: The object to convert.

    Returns:
        float: The float representation of the object.
    """
    return float(obj)


@as_function_node("round")
def Round(obj):
    """
    Rounds a number to the nearest integer.

    Args:
        obj: The number to round.

    Returns:
        int: The rounded number.
    """
    return round(obj)



"""
A base class for nodal objects that have internal structure -- i.e. they hold a
sub-graph
"""

from __future__ import annotations

from abc import ABC
from collections.abc import Callable
from time import sleep
from typing import TYPE_CHECKING

import typeguard
from pyiron_snippets.colors import SeabornColors
from pyiron_snippets.dotdict import DotDict

from pyiron_workflow.create import HasCreator
from pyiron_workflow.mixin.lexical import LexicalParent
from pyiron_workflow.node import Node
from pyiron_workflow.topology import (
    get_nodes_in_data_tree,
    set_run_connections_according_to_dag,
    set_run_connections_according_to_linear_dag,
)

if TYPE_CHECKING:
    from pyiron_workflow.channels import (
        InputSignal,
        OutputSignal,
    )
    from pyiron_workflow.storage import BackendIdentifier, StorageInterface


def _get_graph_as_dict(composite: Composite) -> dict:
    if not isinstance(composite, Composite):
        return composite
    return {
        "object": composite,
        "nodes": {n.full_label: _get_graph_as_dict(n) for n in composite},
        "edges": {
            "data": {
                (out.full_label, inp.full_label): (out, inp)
                for n in composite
                for out in n.outputs
                for inp in out.connections
            },
            "signal": {
                (out.full_label, inp.full_label): (out, inp)
                for n in composite
                for out in n.signals.output
                for inp in out.connections
            },
        },
    }


class FailedChildError(RuntimeError):
    """Raise when one or more child nodes raise exceptions."""


class Composite(LexicalParent[Node], HasCreator, Node, ABC):
    """
    A base class for nodes that have internal graph structure -- i.e. they hold a
    collection of child nodes and their computation is to execute that graph.

    Attributes:
        strict_naming (bool): When true, repeated assignment of a new node to an
         existing node label will raise an error, otherwise the label gets appended
         with an index and the assignment proceeds. (Default is true: disallow assigning
         to existing labels.)
        create (Creator): A tool for adding new nodes to this subgraph.
        provenance_by_completion (list[str]): The child nodes (by label) in the order
            that they completed on the last :meth:`run` call.
        provenance_by_execution (list[str]): The child nodes (by label) in the order
            that they started executing on the last :meth:`run` call.
        running_children (list[str]): The names of children who are currently running.
        signal_queue (list[tuple[OutputSignal, InputSignal]]): Pending signal event
            pairs from child execution flow connections.
        starting_nodes (None | list[pyiron_workflow.node.Node]): A subset
         of the owned nodes to be used on running. Only necessary if the execution graph
         has been manually specified with `run` signals. (Default is an empty list.)
        wrap (Wrappers): A tool for accessing node-creating decorators

    """

    def __init__(
        self,
        *args,
        label: str | None = None,
        parent: Composite | None = None,
        delete_existing_savefiles: bool = False,
        autoload: BackendIdentifier | StorageInterface | None = None,
        autorun: bool = False,
        checkpoint: BackendIdentifier | StorageInterface | None = None,
        strict_naming: bool = True,
        **kwargs,
    ):
        self.starting_nodes: list[Node] = []
        self.provenance_by_execution: list[str] = []
        self.provenance_by_completion: list[str] = []
        self.running_children: list[str] = []
        self.signal_queue: list[tuple[OutputSignal, InputSignal]] = []
        self._child_sleep_interval = 0.01  # How long to wait when the signal_queue is
        # empty but the running_children list is not

        super().__init__(
            *args,
            label=label,
            parent=parent,
            delete_existing_savefiles=delete_existing_savefiles,
            autoload=autoload,
            autorun=autorun,
            checkpoint=checkpoint,
            strict_naming=strict_naming,
            **kwargs,
        )

    @classmethod
    def child_type(cls) -> type[Node]:
        return Node

    def activate_strict_hints(self):
        """Recursively activate strict type hints."""
        super().activate_strict_hints()
        for node in self:
            node.activate_strict_hints()

    def deactivate_strict_hints(self):
        """Recursively de-activate strict type hints."""
        super().deactivate_strict_hints()
        for node in self:
            node.deactivate_strict_hints()

    @property
    def use_cache(self) -> bool:
        """
        Composite nodes determine the cache usage by the cache usage of all children
        recursively.

        Setting this property at the composite level sets it for all children
        recursively.
        """
        return all(c.use_cache for c in self.children.values())

    @use_cache.setter
    def use_cache(self, value: bool):
        for c in self.children.values():
            c.use_cache = value

    @property
    def cache_hit(self) -> bool:
        return not any(c.running for c in self.children.values()) and super().cache_hit

    def _on_cache_miss(self) -> None:
        super()._on_cache_miss()
        # Reset provenance and run status trackers
        self.provenance_by_execution = []
        self.provenance_by_completion = []
        self.running_children = [n.label for n in self if n.running]
        self.signal_queue = []

    def _on_run(self):
        if len(self.running_children) > 0:  # Start from a broken process
            for label in self.running_children:
                if self.children[label]._is_using_wrapped_excutorlib_executor():
                    self.running_children.remove(label)
                self.children[label].run()
                # Running children will find serialized result and proceed,
                # or raise an error because they're already running
        else:  # Start fresh
            for node in self.starting_nodes:
                node.run()

        self._run_while_children_or_signals_exist()

        return self

    def _run_while_children_or_signals_exist(self):
        errors = {}
        while len(self.running_children) > 0 or len(self.signal_queue) > 0:
            try:
                firing, receiving = self.signal_queue.pop(0)
                try:
                    receiving(firing)
                except Exception as e:
                    errors[receiving.full_label] = e
            except IndexError:
                # The signal queue is empty, but there is still someone running...
                sleep(self._child_sleep_interval)

        if len(errors) == 1:
            raise FailedChildError(
                f"{self.full_label} encountered error in child: {errors}"
            ) from next(iter(errors.values()))
        elif len(errors) > 1:
            raise FailedChildError(
                f"{self.full_label} encountered multiple errors in children: {errors}"
            ) from None

    def register_child_starting(self, child: Node) -> None:
        """
        To be called by children when they start their run cycle.

        Args:
            child [Node]: The child that is finished and would like to fire its `ran`
                signal. Should always be a child of `self`, but this is not explicitly
                verified at runtime.
        """
        self.provenance_by_execution.append(child.label)
        self.running_children.append(child.label)

    def register_child_finished(self, child: Node) -> None:
        """
        To be called by children when they are finished their run.

        Args:
            child [Node]: The child that is finished and would like to fire its `ran`
                signal. Should always be a child of `self`, but this is not explicitly
                verified at runtime.
        """
        try:
            self.running_children.remove(child.label)
            self.provenance_by_completion.append(child.label)
        except ValueError as e:
            raise KeyError(
                f"No element {child.label} to remove while {self.running_children}, "
                f"{self.provenance_by_execution}, {self.provenance_by_completion}"
            ) from e

    def register_child_emitting(self, child: Node) -> None:
        """
        To be called by children when they want to emit their signals.

        Args:
            child [Node]: The child that is finished and would like to fire its `ran`
                signal (and possibly others). Should always be a child of `self`, but
                this is not explicitly verified at runtime.
        """
        for firing in child.emitting_channels:
            for receiving in firing.connections:
                self.signal_queue.append((firing, receiving))

    @property
    def run_args(self) -> tuple[tuple, dict]:
        return (), {}

    def process_run_result(self, run_output):
        if run_output is not self:
            self._parse_remotely_executed_self(run_output)
        return self._outputs_to_run_return()

    def _parse_remotely_executed_self(self, other_self):
        # Un-parent existing nodes before ditching them
        for node in self:
            node._parent = None
            node._detached_parent_path = None
        other_self.running = False  # It's done now
        state = self._get_state_from_remote_other(other_self)
        self.__setstate__(state)

    def _get_state_from_remote_other(self, other_self):
        state = other_self.__getstate__()
        state.pop("_executor")  # Got overridden to None for __getstate__, so keep local
        state.pop("_parent")  # Got overridden to None for __getstate__, so keep local
        state.pop("_detached_parent_path")
        return state

    def disconnect_run(self) -> list[tuple[InputSignal, OutputSignal]]:
        """
        Disconnect all `signals.input.run` connections on all child nodes.

        Returns:
            list[tuple[InputSignal, OutputSignal]]: Any disconnected pairs.
        """
        disconnected_pairs = []
        for node in self.children.values():
            disconnected_pairs.extend(node.signals.disconnect_run())
        return disconnected_pairs

    def set_run_signals_to_dag_execution(self):
        """
        Disconnects all `signals.input.run` connections among children and attempts to
        reconnect these according to the DAG flow of the data. On success, sets the
        starting nodes to just be the upstream-most node in this linear DAG flow.
        """
        if len(self.children) > 0:
            _, upstream_most_nodes = set_run_connections_according_to_dag(self.children)
            self.starting_nodes = upstream_most_nodes

    def add_child(
        self,
        child: Node,
        label: str | None = None,
        strict_naming: bool | None = None,
    ) -> Node:
        """Add the node instance to this subgraph."""
        self.clear_cache()  # Reset cache after graph change
        return super().add_child(child, label=label, strict_naming=strict_naming)

    def push_child(self, child: Node | str, *args, **kwargs):
        """
        Run a child node in a "push" configuration.

        Args:
            child (Node|str): The child node to push.
            *args: Additional positional arguments passed to the child node.
            **kwargs: Additional keyword arguments passed to the child node.

        Returns:
            (Any | Future): The result of running the node, or a futures object (if
                running on an executor).
        """
        typeguard.check_type(child, Node | str)

        problem: str | None = None
        if isinstance(child, Node):
            if child.parent is not self:
                problem = child.full_label
            else:
                child_node = child
        elif isinstance(child, str):
            if child not in self.child_labels:
                problem = child
            else:
                child_node = self.children[child]
        if problem is not None:
            raise ValueError(
                f"Child {problem} not found among {self.full_label}'s children: "
                f"{self.child_labels}"
            )
        return child_node.run(*args, **kwargs)

    def remove_child(self, child: Node | str) -> Node:
        """
        Remove a child from the :attr:`children` collection, disconnecting it and
        setting its :attr:`parent` to None.

        Args:
            child (Node|str): The child (or its label) to remove.

        Returns:
            (Node): The (now disconnected and de-parented) (former) child node.
        """
        child = super().remove_child(child)
        child.disconnect()
        if child in self.starting_nodes:
            self.starting_nodes.remove(child)
        self.clear_cache()  # Reset cache after graph change
        return child

    def executor_shutdown(self, wait=True, *, cancel_futures=False):
        """
        Invoke shutdown on the executor (if present), and recursively invoke shutdown
        for children.
        """
        super().executor_shutdown(wait=wait, cancel_futures=cancel_futures)
        for node in self:
            node.executor_shutdown(wait=wait, cancel_futures=cancel_futures)

    def __setattr__(self, key: str, node: Node):
        if isinstance(node, Composite) and key in ["_parent", "parent"]:
            # This is an edge case for assigning a node to an attribute
            super().__setattr__(key, node)
        elif isinstance(node, Node):
            self.add_child(node, label=key)
        else:
            super().__setattr__(key, node)

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    @property
    def color(self) -> str:
        """For drawing the graph"""
        return SeabornColors.brown

    @property
    def graph_as_dict(self) -> dict:
        """
        A nested dictionary representation of the computation graph using full labels
        as keys and objects as values.
        """
        return _get_graph_as_dict(self)

    def _get_connections_as_strings(
        self, panel_getter: Callable
    ) -> list[tuple[tuple[str, str], tuple[str, str]]]:
        """
        Connections between children in string representation based on labels.

        The string representation helps storage, and having it as a property ensures
        the name is protected.
        """
        return [
            ((inp.owner.label, inp.label), (out.owner.label, out.label))
            for child in self
            for inp in panel_getter(child)
            for out in inp.connections
        ]

    @staticmethod
    def _get_data_inputs(node: Node):
        return node.inputs

    @staticmethod
    def _get_signals_input(node: Node):
        return node.signals.input

    @property
    def _child_data_connections(self) -> list[tuple[tuple[str, str], tuple[str, str]]]:
        return self._get_connections_as_strings(self._get_data_inputs)

    @property
    def _child_signal_connections(
        self,
    ) -> list[tuple[tuple[str, str], tuple[str, str]]]:
        return self._get_connections_as_strings(self._get_signals_input)

    @property
    def _starting_node_labels(self):
        # As a property so it appears in `__dir__` and thus is guaranteed to not
        # conflict with a child node name in the state
        return tuple(n.label for n in self.starting_nodes)

    def __getstate__(self):
        state = super().__getstate__()
        # Store connections as strings
        state["_child_data_connections"] = self._child_data_connections
        state["_child_signal_connections"] = self._child_signal_connections

        # Also remove the starting node instances
        del state["starting_nodes"]
        state["_starting_node_labels"] = self._starting_node_labels

        return state

    def __setstate__(self, state):
        # Purge child connection info from the state
        child_data_connections = state.pop("_child_data_connections")
        child_signal_connections = state.pop("_child_signal_connections")
        # Restore starting nodes
        state["starting_nodes"] = [
            state[label] for label in state.pop("_starting_node_labels")
        ]

        super().__setstate__(state)

        # Nodes don't store connection information, so restore it to them
        self._restore_data_connections_from_strings(child_data_connections)
        self._restore_signal_connections_from_strings(child_signal_connections)

    @staticmethod
    def _restore_connections_from_strings(
        nodes: dict[str, Node] | DotDict[str, Node],
        connections: list[tuple[tuple[str, str], tuple[str, str]]],
        input_panel_getter: Callable,
        output_panel_getter: Callable,
    ) -> None:
        """
        Set connections among a dictionary of nodes.

        This is useful for recreating node connections after (de)serialization of the
        individual nodes, which don't know about their connections (that information is
        the responsibility of their parent `Composite`).

        Args:
            nodes (dict[Node]): The nodes to connect.
            connections (list[tuple[tuple[str, str], tuple[str, str]]]): Connections
                among these nodes in the format ((input node label, input channel label
                ), (output node label, output channel label)).
        """
        for (inp_node, inp), (out_node, out) in connections:
            input_panel_getter(nodes[inp_node])[inp].connect(
                output_panel_getter(nodes[out_node])[out]
            )

    @staticmethod
    def _get_data_outputs(node: Node):
        return node.outputs

    @staticmethod
    def _get_signals_output(node: Node):
        return node.signals.output

    def _restore_data_connections_from_strings(
        self, connections: list[tuple[tuple[str, str], tuple[str, str]]]
    ) -> None:
        self._restore_connections_from_strings(
            self.children,
            connections,
            self._get_data_inputs,
            self._get_data_outputs,
        )

    def _restore_signal_connections_from_strings(
        self, connections: list[tuple[tuple[str, str], tuple[str, str]]]
    ) -> None:
        self._restore_connections_from_strings(
            self.children,
            connections,
            self._get_signals_input,
            self._get_signals_output,
        )

    @property
    def import_ready(self) -> bool:
        return super().import_ready and all(node.import_ready for node in self)

    def report_import_readiness(self, tabs=0, report_so_far=""):
        report = super().report_import_readiness(tabs=tabs, report_so_far=report_so_far)
        for node in self:
            report = node.report_import_readiness(tabs=tabs + 1, report_so_far=report)
        return report

    def run_data_tree_for_child(self, node: Node) -> None:
        """
        Use topological analysis to build a tree of all upstream dependencies and run
        them.

        This method is called by a child node when it needs to run its data tree and has
        a parent. The parent (this composite) handles the execution of the data tree.

        Args:
            node (Node): The child node that initiated the data tree run.
        """

        data_tree_nodes = get_nodes_in_data_tree(node)
        for n in data_tree_nodes:
            if n.executor is not None:
                raise ValueError(
                    f"Running the data tree is pull-paradigm action, and is "
                    f"incompatible with using executors. While running "
                    f"{node.full_label}, an executor request was found on "
                    f"{n.full_label}"
                )

        nodes = {n.label: n for n in data_tree_nodes}

        disconnected_pairs, starters = set_run_connections_according_to_linear_dag(
            nodes
        )
        data_tree_starters = list(set(starters).intersection(data_tree_nodes))

        original_starting_nodes = self.starting_nodes
        # We need these for state recovery later, even if we crash

        try:
            if len(data_tree_starters) > 1 or data_tree_starters[0] is not node:
                node.signals.disconnect_run()
                # Don't let anything upstream trigger _this_ node

                self.starting_nodes = data_tree_starters
                self.run()
            # Otherwise the requested node is the only one in the data tree, so there's
            # nothing upstream to run.
        finally:
            # No matter what, restore the original connections and labels afterwards
            for n in nodes.values():
                n.signals.disconnect_run()
            for c1, c2 in disconnected_pairs:
                c1.connect(c2)
            self.starting_nodes = original_starting_nodes



"""
A submodule to hold classes that inherit from :class:`Node`.

I.e. downstream stuff relative to :class:`Node`.
Includes factory functions for such classes, etc.

:class:`Workflow` is exempted from this directory and kept at the main level due to its
importance as the entry point for users.
"""
